/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * A simulator that simulates the action workload scheduling process   
 * within the shared {@link Photo} or {@link Warn} action operator in Aorta.                
 *                                                         
 * It is developed for the sake of convenient performance evaluation  
 * and comparison of different scheduling algorithms. 
 */ 
public class AortaSchedulingSimulator {
	
	public static final byte UNIFORM = 0x01;
	public static final byte SKEWED = 0x02;
	public static final byte ZIPFIAN = 0x03;
	public static final byte UNIFORM_TC = 0x04;
	public static final byte ZIPFIAN_TC = 0x05;
	
	public static final byte CAMERA = 0x01;
	public static final byte ROBOT = 0x02;
	
	byte distribution;
	
	byte deviceType;
	
	int deviceNum;
	int requestNum;
	double skewness;
	
	
	/**
	 * @param distribution the scheduling workload distribution, either {@link #UNIFORM} or {@link #SKEWED}
	 * @param deviceNum the number of cameras involved in the scheduling
	 * @param requestNum the number of action requets contained in the workload
	 * @param skewness for skewed workload only, see {@link SkewedSyntheticWorkload#skewness}
	 */
	public AortaSchedulingSimulator (byte distribution, byte deviceType, int deviceNum, int requestNum, double skewness)
	  {
	  	this.distribution = distribution;
	  	this.deviceType = deviceType;
	  	this.deviceNum = deviceNum;
	  	this.requestNum = requestNum;
	  	this.skewness = skewness;
	  }
	
	
	/**
	 * Generate a synthetic workload with uniform or skewed workload distribution and then
	 * perform the five scheduling algorithms we study on the workload generated. 
	 *
	 * @see ActionRequest
	 * @see UniformSyntheticWorkload
	 * @see SkewedSyntheticWorkload
	 * @see LFRFA
	 * @see SRFE
	 * @see SRFAE
	 * @see SA
	 * @see LIST
	 * @see RANDOM
	 */
	public void schedule ()
	  {	
		long startTime, endTime;
		String logFile;
		Vector requests = null, requests_backup = null;
		PrintWriter logFileWriter = null;
		
 					
		System.out.println("\t\t     Aorta Action Workload Scheduling Simulator");
		System.out.println("A grouping time interval begins.");
		  
		if (distribution == UNIFORM)
		  {
		  	UniformSyntheticWorkload usw = new UniformSyntheticWorkload(deviceType, deviceNum, requestNum);
	        requests = usw.generate();
	     }
	    else if (distribution == SKEWED)
	           {
	      	     SkewedSyntheticWorkload ssw = new SkewedSyntheticWorkload(deviceType, deviceNum, requestNum, skewness);
	             requests = ssw.generate();	
	           }
	         else if (distribution == ZIPFIAN)
	                {
	                  ZipfianSyntheticWorkload zsw = new ZipfianSyntheticWorkload(deviceType, deviceNum, requestNum);
	                  requests = zsw.generate();	
	                }
	               else if (distribution == UNIFORM_TC)
	                      {
	                        UniformTCSyntheticWorkload utcsw = new UniformTCSyntheticWorkload(deviceType, deviceNum, requestNum);
	                        requests = utcsw.generate();	
	                      }
	                    else if (distribution == ZIPFIAN_TC)
	                           {
	                             ZipfianTCSyntheticWorkload ztcsw = new ZipfianTCSyntheticWorkload(deviceType, deviceNum, requestNum);
	                             requests = ztcsw.generate();	
	                           }
	                         else
	                           {
	           	                 System.err.println("Invalid workload distribution specification.");
	           	                 System.exit(1);
	                           }
	                 
	    /* for (int i=0;i<requests.size();i++)
		  {
		  	System.out.print("aid = " + ((ActionRequest)requests.elementAt(i)).aid + "   ");
		  	System.out.print("deadline = " + ((ActionRequest)requests.elementAt(i)).deadline + " ");
		  	System.out.print("didVector = (");
		  	for (int j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
		  	  System.out.print(((ActionRequest)requests.elementAt(i)).didVector.elementAt(j) + ",");
		  	System.out.print(")   ");
		  	System.out.print("costVector = (");
		  	for (int j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
		  	  System.out.print(((ActionRequest)requests.elementAt(i)).costVector.elementAt(j) + ",");
		  	System.out.print(")   ");		  	
		  	System.out.println();
          } 
        System.exit(0); */
		/* for (int i=0;i<requests.size();i++)
		  {
		  	System.out.println("Request #" + (i+1) +".");
		  	for (int j=0;j< ((ActionRequest)requests.elementAt(i)).params.size();j++)
		    {
		      Tuple t = (Tuple)((ActionRequest)requests.elementAt(i)).params.elementAt(j);
		      for (int k=0;k< t.fields.size();k++)
		        System.out.print(t.fields.elementAt(k) + "   ");
		      System.out.println();
		      System.out.println();
		    } }
		System.exit(0);*/
	                       
	    // LFRFA + SRFE
		System.out.println("LFRFA + SRFE ... ");
		logFile = "../logs/scheduling_logs/LFRFA+SRFE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests,requests_backup);  // backup the workload
		startTime = System.currentTimeMillis();
	    LFRFA lfrfa = new LFRFA (requests);
		lfrfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LFRFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		SRFE srfe = new SRFE (deviceType,requests,logFileWriter,endTime - startTime);
		srfe.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();  	  
			  
		/* // LFRFA + LRFE
		System.out.println("LFRFA + LRFE ... ");
		logFile = "../logs/scheduling_logs/LFRFA+LRFE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests,requests_backup);
		startTime = System.currentTimeMillis();
	    lfrfa = new LFRFA (requests);
		lfrfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LFRFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		LRFE lrfe = new LRFE (deviceType,requests,logFileWriter,endTime - startTime);
		lrfe.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();	   	  
			  		   	  
		
		// LFRFA + RRE
		System.out.println("LFRFA + RRE ... ");
		logFile = "../logs/scheduling_logs/LFRFA+RRE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests,requests_backup);
		startTime = System.currentTimeMillis();
	    lfrfa = new LFRFA (requests);
		lfrfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LFRFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		RRE rre = new RRE (deviceType,requests,logFileWriter,endTime - startTime);
		rre.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
		   	  
			  		   	  
		// SRFAE
		System.out.println("SRFAE ... ");
		logFile = "../logs/scheduling_logs/SRFAE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests,requests_backup);
		SRFAE srfae = new SRFAE (deviceType,requests,logFileWriter);
		srfae.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
			 
		  
		/* // LRFAE 
		System.out.println("LRFAE ... ");
		logFile = "../logs/scheduling_logs/LRFAE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LRFAE lrfae = new LRFAE (deviceType,requests,logFileWriter);
		lrfae.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
		
		
		// SA
		System.out.println("SA ... ");
		logFile = "../logs/scheduling_logs/SA.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println(); 
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		SA sa = new SA (deviceType,requests,logFileWriter);
		sa.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
			   
			
		// LFM-LFJ
		System.out.println("LFM-LFJ ... ");
		logFile = "../logs/scheduling_logs/LFM-LFJ.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LFM_LFJ lfm_lfj = new LFM_LFJ (deviceType,requests,logFileWriter);
		lfm_lfj.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
		
		
		/* // LIST
		System.out.println("LIST ... ");
		logFile = "../logs/scheduling_logs/LIST.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LIST list = new LIST (deviceType,requests,logFileWriter);
		list.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();	
				  
		
		// LJF
		System.out.println("LJF ... ");
		logFile = "../logs/scheduling_logs/LJF.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LJF ljf = new LJF (deviceType,requests,logFileWriter);
		ljf.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();	
		
			    
		// SJF
		System.out.println("SJF ... ");
		logFile = "../logs/scheduling_logs/SJF.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		SJF sjf = new SJF (deviceType,requests,logFileWriter);
		sjf.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
				   	
			
		// RANDOM
	    System.out.println("RANDOM ... ");
		logFile = "../logs/scheduling_logs/RANDOM.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		RANDOM random = new RANDOM (deviceType,requests,logFileWriter);
		random.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
				  
		
        System.out.println("The grouping time interval ends.");       
      }  
      
      
    /** 
     * Backup the generated synthetic workload so that the same workload can be used for all five scheduling algorithm.
     * 
     * @param requests the scheduling workload for backup
     * @param requests_backup a backup of the workload
     */
    public void backupRequests (Vector requests, Vector requests_backup)
      {
      	int i;
      	for (i=0;i<requests.size();i++)
		  requests_backup.addElement((ActionRequest)requests.elementAt(i));
      }
	
	
	/** Specify the parameters of the scheduling simulator and start the simulator. */
	public static void main (String argv [])  
	  {
	  	// change these parameters in different experiments
	  	byte distribution = AortaSchedulingSimulator.UNIFORM_TC;
	  	
	  	byte deviceType = AortaSchedulingSimulator.CAMERA;
	  	
	  	int deviceNum = 10;
	  	
	  	int requestNum = 20;
	  	
	  	double skewness = 1;
	  	
	  	AortaSchedulingSimulator aortass = new AortaSchedulingSimulator(distribution,deviceType, deviceNum,requestNum,skewness);
	  	
	  	aortass.schedule();
	  }	
} 