/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the random scheduling.
 *
 * @see AortaSchedulingSimulator#schedule
 * @see SinglePassThread#scheduleCameras
 * @see SinglePassThread#scheduleRobots
 */
public class RANDOM {
	
	byte deviceType;
	
	Vector requests;
	
	PrintWriter logFileWriter;
	
	deviceWorkload allDeviceWorkload [];
	
	double singleServiceTime, singleSchedulingTime;
 	double overallServiceTime, overallSchedulingTime;
 	

	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 *  
	 * @see ActionRequest
	 */
	public RANDOM (byte deviceType, Vector requests, PrintWriter logFileWriter) 
	  {
		this.deviceType = deviceType;
		this.requests = requests;
		this.logFileWriter = logFileWriter;	
		
		overallServiceTime = 0; overallSchedulingTime = 0;
		
		long startTime = System.currentTimeMillis();
		
		initialize ();
		
		assignment();    	
      	
    	long endTime = System.currentTimeMillis();
    	
    	overallSchedulingTime += endTime - startTime;	
	  } 	
	
	public void initialize ()
	  {
	  	int i,j;
	  	
	  	Vector didVector = new Vector (20,10);
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    
    	allDeviceWorkload = new deviceWorkload [didVector.size()];
    	
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	allDeviceWorkload[i] = new deviceWorkload();
    	  	allDeviceWorkload[i].did = (String)didVector.elementAt(i);
    	  }     	
	  }	  
	  
	/** Assign the requests to the candidate devices in a random fashion. */
	public void assignment ()
	  {
	   	int i,j;
		int sel_no;
		
		Random ran = new Random();
		ActionRequest ar;
		
		// assign the requests to devices in a pure random fashion
    	for (i=0;i<requests.size();i++)
    	  {
    	   	ar = (ActionRequest)requests.elementAt(i);
    	   	
    	   	sel_no = Math.abs(ran.nextInt()) % ar.didVector.size();
    	  	
    	  	String did = (String)ar.didVector.elementAt(sel_no);
    	  	
    	  	for (j=0;j<allDeviceWorkload.length;j++)
    	  	  if (allDeviceWorkload[j].did.equals(did))
    	  	    break;
    	  	    
    	  	allDeviceWorkload[j].requests.addElement((ActionRequest)ar);
    	  	
    	  	ar.sel_no = sel_no;    	  	
    	  }	
	  }	  
	  
	/** 
	 * Execute the requests on the candidate devices. 
	 *
	 * @see #singleCameraExecution
 	 * @see #singleRobotExecution
 	 */ 	
	public void execution() 
	  {
		int i;
		double maxSingleServiceTime = 0, maxSingleSchedulingTime = 0;
		 	  	
 	  	for (i=0;i<allDeviceWorkload.length;i++)
 	  	  {
 	  	  	logFileWriter.println("\n#");
 	  	  	logFileWriter.println("Device " + allDeviceWorkload[i].did + " :");
 	  	  	singleDeviceExecution(allDeviceWorkload[i]);
 	  	  	if (singleServiceTime > maxSingleServiceTime) 
 	  	  	  maxSingleServiceTime = singleServiceTime;
 	  	  	if (singleSchedulingTime > maxSingleSchedulingTime)
 	  	  	  maxSingleSchedulingTime = singleSchedulingTime;
 	  	  	logFileWriter.println("#"); 	  	  	 	  	
 	      } 
 	    
 	    overallServiceTime += maxSingleServiceTime; 
 	    overallSchedulingTime += maxSingleSchedulingTime;  
 	    
 	    logFileWriter.println();
 	    logFileWriter.println("#");
 	    logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 	    logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 	    logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 	    logFileWriter.println("#");	    		
	  }
	  
	/** Execute the requests that have been assigned to a single camera in a random fasihon. */
	public void singleDeviceExecution(deviceWorkload dw)
	  {
	  	if (deviceType == AortaSchedulingSimulator.CAMERA)
	  	  singleCameraExecution(dw);
	  	else
	  	  if (deviceType == AortaSchedulingSimulator.ROBOT)
	  	    singleRobotExecution(dw);
	  	  else
	  	    {
      	      logFileWriter.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      logFileWriter.close();
      	      
      	      System.err.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      System.exit(0);     	      
      	    }	  	  
	  }	
	
	public void singleCameraExecution(deviceWorkload dw)  
      {
	    int sel_no;
    	int count = 0;
    	int pan, tilt, zoom;
    	int deltaPan, deltaTilt, deltaZoom;
    	long startTime, endTime;
    	double headMovementCost, actionCost;
    	Random ran = new Random(); 	
    	Tuple t;
    	ActionRequest ar;
    	CameraSimulator camSim;    	
    	
    	camSim = new CameraSimulator(dw.did);    	
    	singleServiceTime = 0; singleSchedulingTime = 0;
    	singleSchedulingTime += camSim.polling();    	
    	while (dw.requests.size() != 0)
    	  {
    	  	logFileWriter.println("Current physical status of Camera " + camSim.id + ": pan = " + camSim.pan + "  tilt = " + camSim.tilt + "  zoom = " + camSim.zoom + ".");
            logFileWriter.println();
             
            startTime = System.currentTimeMillis(); 
               	  	
    	  	if (dw.requests.size() == 1)
    	  	  sel_no = 0;
    	  	else
    	  	  sel_no = Math.abs(ran.nextInt()) % dw.requests.size();    	  	
    	   	
    	  	endTime = System.currentTimeMillis();
    	  	singleSchedulingTime += endTime - startTime;
    	  	    	  	
    	  	ar = (ActionRequest)dw.requests.elementAt(sel_no);
    	  	t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	deltaPan = Math.abs(pan - camSim.pan);
    	   	deltaTilt = Math.abs(tilt - camSim.tilt);
    	   	deltaZoom = Math.abs(zoom - camSim.zoom);
    	   	headMovementCost = camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
    	   	actionCost = estimatePhotoActionCost(headMovementCost); 
    	   	singleServiceTime += actionCost; 
    	   	
    	  	camSim.pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	camSim.tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	camSim.zoom = Integer.parseInt((String)t.fields.elementAt(2)); 
    	  	
    	  	logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced. ");
    	  	logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	logFileWriter.println();
    	  	    	  	 
    	  	dw.requests.removeElementAt(sel_no);  
    	  	
    	  	for (int i=0;i<dw.requests.size();i++)
    	  	  {    	  	  	
    	  	  	ar = (ActionRequest)dw.requests.elementAt(i);  
    	  	  	
    	  	  	if (ar.deadline != Double.MAX_VALUE)
    	  	  	  if (ar.deadline < overallSchedulingTime + singleSchedulingTime + singleServiceTime)
    	  	  	    {
    	  	  	        logFileWriter.println("Request #" + ar.aid + " is timed out and dropped.");
    	  	  	        logFileWriter.println();	
    	  	  	    	dw.requests.removeElementAt(i);
    	  	  	    	i--;
    	  	  	     }    	  	  	        	  	  	
    	  	  }	  	 	  	   	  	
    	  }    	
    	 
    	logFileWriter.println("Service time on this device: " + singleServiceTime + " ms.");
 	  	logFileWriter.println("Scheduling time on this device: " + singleSchedulingTime + " ms."); 	  	 	  	  	  	     	
      } 
     public void singleRobotExecution(deviceWorkload dw) 
	  {
		int count = 0;
    	double angle, actionCost;
    	Tuple t;
    	Point loc = new Point();
    	ActionRequest ar;
    	ER1Simulator er1Sim;   	    	
    
    	er1Sim = new ER1Simulator(dw.did);    	
        singleServiceTime = 0; singleSchedulingTime = 0;
        singleSchedulingTime += ER1Simulator.STANDARD_POLLING_TIME;  	    	
    	while (dw.requests.size() != 0)
    	  {
    	  	logFileWriter.println("Current physical status of Robot " + er1Sim.id + ": loc = (" + er1Sim.loc.x + "," + er1Sim.loc.y + "," + er1Sim.loc.z + ") " + "angle = " + er1Sim.angle + ".");
            logFileWriter.println();
                       
            ar = (ActionRequest)dw.requests.elementAt(0);
    	  	t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	actionCost = estimateWarnActionCost(er1Sim.loc,er1Sim.angle,loc);  
    	   	singleServiceTime += actionCost;
    	   	
    	   	angle = computeTargetAngle(er1Sim.loc,loc);
    	   	er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	er1Sim.setAngle(angle);   	   	
    	   	    	   	
    	  	logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced. ");
    	  	logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	logFileWriter.println();
    	  	    	  	 
    	  	dw.requests.removeElementAt(0); 
    	  	
    	  	for (int i=0;i<dw.requests.size();i++)
    	  	  {    	  	  	
    	  	  	ar = (ActionRequest)dw.requests.elementAt(i);  
    	  	  	
    	  	  	if (ar.deadline != Double.MAX_VALUE)
    	  	  	  if (ar.deadline < overallSchedulingTime + singleSchedulingTime + singleServiceTime)
    	  	  	    {
    	  	  	        logFileWriter.println("Request #" + ar.aid + " is timed out and dropped.");
    	  	  	        logFileWriter.println();	
    	  	  	    	dw.requests.removeElementAt(i);
    	  	  	    	i--;
    	  	  	     }    	  	  	        	  	  	
    	  	  }	 	  	   	  	
    	  }    	
    	 
    	logFileWriter.println("Service time on this robot: " + singleServiceTime + " ms.");
 	  	logFileWriter.println("Scheduling time on this robot: " + singleSchedulingTime + " ms."); 
 	  	
 	  	try {er1Sim.serverSocket.close();}
 	  	catch (Exception e) {e.printStackTrace();}  	  	 	  	  	  	     	
      }        
	  
	/** Estimate the cost of a {@link Photo} action. */
	public double estimatePhotoActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_POLLING_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      } 
    
    /** Estimate the cost of a {@link Warn} action. */
   public double estimateWarnActionCost (Point currentLoc, double currentAngle, Point targetLoc)
     {
       double actionCost = 0;
       actionCost += Math.sqrt(Math.pow(targetLoc.x-currentLoc.x,2)+Math.pow(targetLoc.y-currentLoc.y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       double deltaAngle = computeTargetAngle(currentLoc, targetLoc);
       deltaAngle -= currentAngle;
	   if (deltaAngle > 180)
	     deltaAngle -= 360;
	   if (deltaAngle <= -180)
	     deltaAngle += 360;
	   actionCost += ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle;
       
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }
     
   public double computeTargetAngle(Point currentLoc, Point targetLoc)
     {
       double currentX = currentLoc.x, currentY = currentLoc.y;
       double targetX = targetLoc.x, targetY = targetLoc.y;  	    
       double deltaAngle; 
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     		     	  	
	      }	
	      
	   return deltaAngle;	
     }
 } 