/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of the SA (Simulated Annealing) scheduling algorithm.
 *
 * @see AortaSchedulingSimulator#schedule
 * @see SinglePassThread#scheduleCameras
 * @see SinglePassThread#scheduleRobots
 */
public class SA {
	
	byte deviceType;
	
	Vector requests;
		
	PrintWriter logFileWriter;	
	
	int selNo [], selNoChanged [];
	
	double singleServiceTime, singleSchedulingTime;
	double overallServiceTime, overallSchedulingTime;
	
	deviceWorkload allDeviceWorkload [];
	deviceWorkload allDeviceWorkloadChanged [];
	
	Random ran = new Random();		
		
	/* SA magic numbers */
	public static final double T0 = 50;
	public static final double a = 0.99f;
	public static final int kmax = 1000;
	public static double imax;
	public static final int ks = 20;
	public static double is;
	
	public static final byte RANDOM_GENERATION = 0x01;
	public static final byte INTRA_PAIRWISE_INTERCHANGE = 0x02;
	public static final byte INTRA_REINSERTION = 0x03;
	public static final byte INTER_PAIRWISE_INTERCHANGE = 0x04;
	public static final byte INTER_REINSERTION = 0x05;
	
	public static final double RANDOM_GENERATION_PROBABILITY = 0.1;
	public static final double INTRA_PAIRWISE_INTERCHANGE_PROBABILITY = 0.2;
	public static final double INTRA_REINSERTION_PROBABILITY = 0.2;
	public static final double INTER_PAIRWISE_INTERCHANGE_PROBABILITY = 0.2;
	public static final double INTER_REINSERTION_PROBABILITY = 0.3;
	
	
	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 *
	 * @see ActionRequest
	 */
	public SA (byte deviceType, Vector requests, PrintWriter logFileWriter) 
	  {
		this.deviceType = deviceType;
		this.requests = requests;
		this.logFileWriter = logFileWriter;
		
		overallServiceTime = 0; overallSchedulingTime = 0;
		
		long startTime = System.currentTimeMillis();
		
		initialize ();
		
		assignment();
		
		long endTime = System.currentTimeMillis();
		
		overallSchedulingTime += endTime - startTime;
      }      
      
    public void initialize ()
	  {
	  	int i,j;		
		
		Vector didVector = new Vector (20,10);
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	
    	allDeviceWorkload = new deviceWorkload [didVector.size()];
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	allDeviceWorkload[i] = new deviceWorkload();
    	  	allDeviceWorkload[i].did = (String)didVector.elementAt(i);   	  	  	 	  	
    	  }    	  	   	
	  }	  
	  
	/** Assign the requests to the candidate devices with a simulated annealing process. */
	public void assignment ()
      {	
		int i;
		long startTime, endTime;
		
		selNo = new int [requests.size()];	
		
		imax = (double)(15*Math.pow(requests.size(),2));
		is = 0.1f*imax;		
    	
    	// start the simulated annealing process, refine the current schedule
    	logFileWriter.println("#");
    	logFileWriter.println("The simulated annealing process begins.");
    	
    	startTime = System.currentTimeMillis();
    	
    	int count = 0;
	    randomGeneration(allDeviceWorkload, selNo); 
	    int k = 1; 
	    double T = T0; 
	    int ks_count = 0;
	    while ( (k <= kmax) && (ks_count < ks) )
	      {
	      	int I = 1; int is_count = 0;
	      	while ( (I <= imax) && (is_count < is) )
	      	  {
	      	  	double prob = Math.abs(ran.nextDouble());
	      	  	byte transitionType;
	      	  	if ( prob <= RANDOM_GENERATION_PROBABILITY )
	      	  	  transitionType = RANDOM_GENERATION;
	      	  	else
	      	  	  if ( prob <= (RANDOM_GENERATION_PROBABILITY + INTRA_PAIRWISE_INTERCHANGE_PROBABILITY) )
	      	  	    transitionType = INTRA_PAIRWISE_INTERCHANGE;
	      	  	  else
	      	  	    if ( prob <= (RANDOM_GENERATION_PROBABILITY + INTRA_PAIRWISE_INTERCHANGE_PROBABILITY + INTRA_REINSERTION_PROBABILITY) )
	      	  	      transitionType = INTRA_REINSERTION;
	      	  	    else
	      	  	      if ( prob <= (RANDOM_GENERATION_PROBABILITY + INTRA_PAIRWISE_INTERCHANGE_PROBABILITY + INTRA_REINSERTION_PROBABILITY + INTER_PAIRWISE_INTERCHANGE_PROBABILITY) )
	      	  	        transitionType = INTER_PAIRWISE_INTERCHANGE;
	      	  	      else
	      	  	        transitionType = INTER_REINSERTION;
	      	  	        
	      	  	newSchedule(transitionType);  
	      	  	double makespan_old = computeMakeSpan(allDeviceWorkload, false);
	      	  	double makespan_new = computeMakeSpan(allDeviceWorkloadChanged, true);
	      	  	if (makespan_new < makespan_old)
	      	  	  {
	      	  	  	
	      	  	  	allDeviceWorkload = allDeviceWorkloadChanged;
	      	  	  	selNo = selNoChanged;
	      	  	  	System.gc();
	      	  	  	is_count ++;
	      	  	  	
	      	  	  	logFileWriter.println("Transition " + count + ": The makespan of the schedule switches from " + makespan_old + " to " + makespan_new + ".");
	      	  	  	// System.out.println("Transition " + count + ": The makespan of the schedule switches from " + makespan_old + " to " + makespan_new + ".");
	      	  	  	ks_count = 0;
	      	  	  }
	      	  	else
	      	  	  {
	      	  	  	double delta_makespan = makespan_old - makespan_new;
	      	  	  	prob = Math.pow(Math.E,delta_makespan/T);
	      	  	  	
	      	  	  	if ( (prob < 1) && (Math.abs(ran.nextDouble()) <= prob) )
	      	  	  	  {
	      	  	  	    allDeviceWorkload = allDeviceWorkloadChanged;
	      	  	  	    selNo = selNoChanged;
	      	  	  	  	
	      	  	  	  	logFileWriter.println("Transition " + count + ": The makespan of the schedule switches from " + makespan_old + " to " + makespan_new + ".");
	      	  	  	  	// System.out.println("Transition " + count + ": The makespan of the schedule switches from " + makespan_old + " to " + makespan_new + ".");
	      	  	  	  	
	      	  	  	  	System.gc();
	      	  	  	    is_count ++;
	      	  	  	    ks_count = 0;	      	  	  	    
	      	  	  	   }
	      	  	  	else
	      	  	  	  {
	      	  	  	  	ks_count ++;
	      	  	  	  	// System.out.println("The makespan does not switch in this round.");
	      	  	  	  }	      	  	  	
	      	  	  }     	  	
	      	  	
	      	  	I++;
	      	  	
	      	  }  // while
	      	  
	      	k++;
	      	T *= a;	     		
	      	
	      }  // while	    
	    
	    endTime = System.currentTimeMillis();
	    
	    logFileWriter.println("Simulated Annealing time: " + (endTime - startTime) + " ms.");
	    logFileWriter.println("The simulated annealing process ends.");
	    logFileWriter.println("#");
    	logFileWriter.println("");    	   		
		
		allDeviceWorkloadChanged = null;
		System.gc();		
		
		for (i=0;i<requests.size();i++)
		  ((ActionRequest)requests.elementAt(i)).sel_no = selNo[i];					
	  }	
	
	/** Randomly genreate a schedule. */
	public void randomGeneration (deviceWorkload [] aw, int [] selNo) 
	  {		
		int i,j;
		int sel_no;
		
		// assign the requests to devices in a pure random fashion
    	for (i=0;i<requests.size();i++)
    	  {
    	  	sel_no = Math.abs(ran.nextInt()) % (((ActionRequest)requests.elementAt(i)).didVector.size() );
    	  	
    	  	String did = (String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(sel_no);
    	  	
    	  	for (j=0;j<aw.length;j++)
    	  	  if (aw[j].did.equals(did))
    	  	    break;
    	  	    
    	  	aw[j].requests.addElement((ActionRequest)requests.elementAt(i));
    	  	
    	  	selNo[i] = sel_no;    	  	
    	  } 	
	  } 	
	
	/** Switch from the current schedul to a new scheule. */
	public void newSchedule (byte transitionType) 
	  {
		int i,j;
		
		allDeviceWorkloadChanged =new deviceWorkload [allDeviceWorkload.length];
		for (i=0;i<allDeviceWorkload.length;i++)
    	  {
    	  	allDeviceWorkloadChanged[i] = new deviceWorkload();
    	  	allDeviceWorkloadChanged[i].did = allDeviceWorkload[i].did;    	  	 	  	
    	  }
    	for (i=0;i<allDeviceWorkload.length;i++)
    	  for (j=0;j<allDeviceWorkload[i].requests.size();j++)
    		allDeviceWorkloadChanged[i].requests.addElement((ActionRequest)allDeviceWorkload[i].requests.elementAt(j));
    	
    	selNoChanged = new int [requests.size()];
    	for (i=0;i<requests.size();i++)
    	  selNoChanged[i] = selNo[i];
    	
    	    	
    	switch (transitionType) 
    	  {
    		case RANDOM_GENERATION:
    		  {
    		  	randomGeneration(allDeviceWorkloadChanged, selNoChanged);
    		  }
    		break;
    		  
    		case INTRA_PAIRWISE_INTERCHANGE:
    		  {
    		  	int device_no = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	while (allDeviceWorkloadChanged[device_no].requests.size() == 0)
    		  	  device_no = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	int request_no1, request_no2;
    		  	if ( allDeviceWorkloadChanged[device_no].requests.size() == 1 )
    		  	  request_no1 = request_no2 = 0;
    		  	else
    		  	  {
    		  	  	request_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no].requests.size();
    		  	    request_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no].requests.size();
    		  	  }
    		  	ActionRequest ar1 = (ActionRequest)allDeviceWorkloadChanged[device_no].requests.elementAt(request_no1);
    		  	ActionRequest ar2 = (ActionRequest)allDeviceWorkloadChanged[device_no].requests.elementAt(request_no2);
    		  	allDeviceWorkloadChanged[device_no].requests.removeElementAt(request_no1);
    		  	allDeviceWorkloadChanged[device_no].requests.insertElementAt(ar2, request_no1);
    		  	allDeviceWorkloadChanged[device_no].requests.removeElementAt(request_no2);
    		  	allDeviceWorkloadChanged[device_no].requests.insertElementAt(ar1, request_no2);
    		  }
    		break;
    		  
    		case INTRA_REINSERTION:
    		  {
    		  	int device_no = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	while (allDeviceWorkloadChanged[device_no].requests.size() == 0)
    		  	  device_no = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	int request_no;
    		  	if ( allDeviceWorkloadChanged[device_no].requests.size() == 1 )
    		  	  request_no = 0;
    		  	else
    		  	  request_no = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no].requests.size();
    		  	ActionRequest ar = (ActionRequest)allDeviceWorkloadChanged[device_no].requests.elementAt(request_no);
    		  	allDeviceWorkloadChanged[device_no].requests.removeElementAt(request_no);
    		  	allDeviceWorkloadChanged[device_no].requests.addElement(ar);		  	    
    		  }
    		  break;
    		  
    		case INTER_PAIRWISE_INTERCHANGE:
    		  {
    		  	int device_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	while (allDeviceWorkloadChanged[device_no1].requests.size() == 0)
    		  	  device_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	int device_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	while (allDeviceWorkloadChanged[device_no2].requests.size() == 0)
    		  	  device_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	  
    		  	int request_no1, request_no2;
    		  	if ( allDeviceWorkloadChanged[device_no1].requests.size() == 1 )
    		  	  request_no1 = 0;
    		  	else
    		  	  request_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no1].requests.size();
    		  	if ( allDeviceWorkloadChanged[device_no2].requests.size() == 1 )
    		  	  request_no2 = 0;
    		  	else
    		  	  request_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no2].requests.size();
    		  	
    		  	ActionRequest ar1 = (ActionRequest)allDeviceWorkloadChanged[device_no1].requests.elementAt(request_no1);
    		  	ActionRequest ar2 = (ActionRequest)allDeviceWorkloadChanged[device_no2].requests.elementAt(request_no2);
    		  	allDeviceWorkloadChanged[device_no1].requests.removeElementAt(request_no1);
    		  	allDeviceWorkloadChanged[device_no1].requests.insertElementAt(ar2, request_no1);
    		  	allDeviceWorkloadChanged[device_no2].requests.removeElementAt(request_no2);
    		  	allDeviceWorkloadChanged[device_no2].requests.insertElementAt(ar1, request_no2); 
    		  	
    		  	int k;
    		  	for (k=0;k<ar1.didVector.size();k++)
    		  	  if (allDeviceWorkloadChanged[device_no2].did.equals((String)ar1.didVector.elementAt(k)))
    		  	    break;
    		    if (k != ar1.didVector.size())
    		  	  selNoChanged[Integer.parseInt(ar1.aid)-1] = k;
    		  	else
    		  	  {
    		  	  	allDeviceWorkloadChanged = allDeviceWorkload;
    		  	  	selNoChanged = selNo;
    		  	  	System.gc();
    		  	    return;
    		  	   }  
    		  	for (k=0;k<ar2.didVector.size();k++)
    		  	  if (allDeviceWorkloadChanged[device_no1].did.equals((String)ar2.didVector.elementAt(k)))
    		  	    break;
    		    if (k != ar2.didVector.size())
    		  	  selNoChanged[Integer.parseInt(ar2.aid)-1] = k;
    		  	else
    		  	  {
    		  	  	allDeviceWorkloadChanged = allDeviceWorkload;
    		  	  	selNoChanged = selNo;
    		  	  	System.gc();
    		  	    return;
    		  	   }     		  	   		  	    
    		  }
    	    break;
            
            case INTER_REINSERTION:
              {
              	int device_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
                while (allDeviceWorkloadChanged[device_no1].requests.size() == 0)
    		  	  device_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;    		  	  
    		  	int device_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged.length;
    		  	
    		  	int request_no1, request_no2;
    		  	if (allDeviceWorkloadChanged[device_no1].requests.size() == 1)
    		  	  request_no1 = 0;
    		  	else
    		  	  request_no1 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no1].requests.size();
    		  	if (allDeviceWorkloadChanged[device_no2].requests.size() <= 1)
    		  	  request_no2 = 0;
    		  	else
    		  	  request_no2 = Math.abs(ran.nextInt()) % allDeviceWorkloadChanged[device_no2].requests.size();
    		  	
    		  	ActionRequest ar = (ActionRequest)allDeviceWorkloadChanged[device_no1].requests.elementAt(request_no1);
    		  	allDeviceWorkloadChanged[device_no1].requests.removeElementAt(request_no1);
    		  	allDeviceWorkloadChanged[device_no2].requests.insertElementAt(ar,request_no2);
    		  	int k;
    		  	for(k=0;k<ar.didVector.size();k++)
    		  	  if (allDeviceWorkloadChanged[device_no2].did.equals((String)ar.didVector.elementAt(k)))
    		  	    break;
    		    if (k != ar.didVector.size())
    		  	  selNoChanged[Integer.parseInt(ar.aid)-1] = k;
    		  	else
    		  	  {
    		  	  	allDeviceWorkloadChanged = allDeviceWorkload;
    		  	  	selNoChanged = selNo;
    		  	  	System.gc();
    		  	    return;
    		  	  }    		  		  	    
              }       		
    		
    	  } // switch	
	  } 	
	
	/** Compute the makespan of a schedules. */
	public double computeMakeSpan(deviceWorkload [] aw, boolean Changed)
	  {
	  	if (deviceType == AortaSchedulingSimulator.CAMERA)
	  	  return computePhotoMakeSpan(aw, Changed);
	  	else
	  	  if (deviceType == AortaSchedulingSimulator.ROBOT)
	  	    return computeWarnMakeSpan(aw, Changed);
	  	  else
	  	    {
      	      logFileWriter.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      logFileWriter.close();
      	      
      	      System.err.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      System.exit(0);
      	      
      	      return Double.MAX_VALUE;
      	    }
	  }	
	  
	public double computePhotoMakeSpan(deviceWorkload [] aw, boolean Changed)
	  {
	  	int i,j;
	  	int pan,tilt,zoom;
	  	int deltaPan, deltaTilt, deltaZoom;
	  	double makespan = Double.MIN_VALUE, makespan_t;
	  	double headMovementCost, actionCost;
	  	Tuple t;
	  	ActionRequest ar;
	  	CameraSimulator camSim;	  	
	  	
	  	for (i=0;i<aw.length;i++)
	  	  {
	  	  	makespan_t = 0;
	  	  	
	  	  	camSim = new CameraSimulator(aw[i].did);
	  	  	
	  	  	for (j=0;j<aw[i].requests.size();j++)
	  	      {
	  	      	ar = (ActionRequest)aw[i].requests.elementAt(j);
	  	      	if (Changed)
	  	      	  t = (Tuple)ar.statusVector.elementAt(selNoChanged[Integer.parseInt(ar.aid)-1]);
	  	      	else
	  	      	  t = (Tuple)ar.statusVector.elementAt(selNo[Integer.parseInt(ar.aid)-1]);
	  	      	
	  	      	pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	    tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	    zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	    deltaPan = Math.abs(pan - camSim.pan);
    	   	    deltaTilt = Math.abs(tilt - camSim.tilt);
    	   	    deltaZoom = Math.abs(zoom - camSim.zoom);
    	   	    headMovementCost = camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
    	   	    actionCost = estimatePhotoActionCost(headMovementCost); 
    	   	    camSim.pan = pan;
    	   	    camSim.tilt = tilt;
    	   	    camSim.zoom = zoom;
    	   	  
	  	      	makespan_t += actionCost;  	      	    	        
	  	      }
	  	      
	  	    if (makespan < makespan_t)
	  	      makespan = makespan_t;	  	      
	  	  }  	
	  	
	  	return makespan;
	  }
	  
		public double computeWarnMakeSpan(deviceWorkload [] aw, boolean Changed)
	  {
	  	int i,j;
	  	double makespan = Double.MIN_VALUE, makespan_t;
	  	double angle, actionCost;
	  	Tuple t;
	  	Point loc = new Point();
	  	ActionRequest ar;
	  	ER1Simulator er1Sim;  	
	  	
	  	for (i=0;i<aw.length;i++)
	  	  {
	  	  	makespan_t = 0;
	  	  	
	  	  	er1Sim = new ER1Simulator(aw[i].did);
	  	  	
	  	  	for (j=0;j<aw[i].requests.size();j++)
	  	      {
	  	      	ar = (ActionRequest)aw[i].requests.elementAt(j);
	  	      	if (Changed)
	  	      	  t = (Tuple)ar.statusVector.elementAt(selNoChanged[Integer.parseInt(ar.aid)-1]);
	  	      	else
	  	      	  t = (Tuple)ar.statusVector.elementAt(selNo[Integer.parseInt(ar.aid)-1]);
	  	      	
	  	      	loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	    loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	    loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	    actionCost = estimateWarnActionCost(er1Sim.loc,er1Sim.angle,loc); 
    	   	    angle = computeTargetAngle(er1Sim.loc,loc);
    	   	    er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	    er1Sim.setAngle(angle);   
    	   	  
	  	      	makespan_t += actionCost;  	      	    	        
	  	      }
	  	      
	  	    if (makespan < makespan_t)
	  	      makespan = makespan_t;
	  	      
	  	    try {er1Sim.serverSocket.close();}
 	  	    catch (Exception e) {e.printStackTrace();}  	  	      
	  	  } 	  	
	  	
	  	return makespan;
	  }	
	
	/** Execute the requests on the candidate devices. */
	public void execution() 
	  {
		int i;
		double maxSingleServiceTime = 0, maxSingleSchedulingTime = 0; 	 
		
	  	for (i=0;i<allDeviceWorkload.length;i++)
 	  	  {
 	  	  	logFileWriter.println("\n#");
 	  	  	logFileWriter.println("Device " + allDeviceWorkload[i].did + " :");
 	  	  	singleDeviceExecution(allDeviceWorkload[i]);
 	  	  	if (singleServiceTime > maxSingleServiceTime) 
 	  	  	  maxSingleServiceTime = singleServiceTime;
 	  	  	if (singleSchedulingTime > maxSingleSchedulingTime)
 	  	  	  maxSingleSchedulingTime = singleSchedulingTime;
 	  	  	logFileWriter.println("#"); 	  	  	 	  	
 	      } 
 	     	    
 	    overallServiceTime += maxSingleServiceTime;
 	    overallSchedulingTime += maxSingleSchedulingTime;
 	    
 	    logFileWriter.println();
 	    logFileWriter.println("#");
 	    logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 	    logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 	    logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 	    logFileWriter.println("#"); 		
      } 	
	
	/** 
	 * Execute the requests that have been assigned to a single device in the FIFO order. 
	 * @see #singleCameraExecution
 	 * @see #singleRobotExecution
 	 */ 
	public void singleDeviceExecution(deviceWorkload dw)
	  {
	  	if (deviceType == AortaSchedulingSimulator.CAMERA)
	  	  singleCameraExecution(dw);
	  	else
	  	  singleRobotExecution(dw);
	  }
	  
	public void singleCameraExecution(deviceWorkload dw) 
	  {
		int count = 0;
    	int pan, tilt, zoom;
    	int deltaPan, deltaTilt, deltaZoom;
    	double headMovementCost, actionCost;
    	Tuple t;
    	ActionRequest ar;
    	CameraSimulator camSim;    	    	
    
    	camSim = new CameraSimulator(dw.did);    	
        singleServiceTime = 0; singleSchedulingTime = 0;
        singleSchedulingTime += camSim.polling();  	    	
    	while (dw.requests.size() != 0)
    	  {
    	  	logFileWriter.println("Current physical status of Camera " + camSim.id + ": pan = " + camSim.pan + "  tilt = " + camSim.tilt + "  zoom = " + camSim.zoom + ".");
            logFileWriter.println();
                       
            ar = (ActionRequest)dw.requests.elementAt(0);
    	  	t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	deltaPan = Math.abs(pan - camSim.pan);
    	   	deltaTilt = Math.abs(tilt - camSim.tilt);
    	   	deltaZoom = Math.abs(zoom - camSim.zoom);
    	   	headMovementCost = camSim.computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
    	   	actionCost = estimatePhotoActionCost(headMovementCost); 
    	   	singleServiceTime += actionCost;
    	   	
    	   	camSim.pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	camSim.tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	camSim.zoom = Integer.parseInt((String)t.fields.elementAt(2));    	   	
    	   	    	   	
    	  	logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced. ");
    	  	logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	logFileWriter.println();
    	  	    	  	 
    	  	dw.requests.removeElementAt(0); 
    	  	
    	  	for (int i=0;i<dw.requests.size();i++)
    	  	  {    	  	  	
    	  	  	ar = (ActionRequest)dw.requests.elementAt(i);  
    	  	  	
    	  	  	if (ar.deadline != Double.MAX_VALUE)
    	  	  	  if (ar.deadline < overallSchedulingTime + singleSchedulingTime + singleServiceTime)
    	  	  	    {
    	  	  	        logFileWriter.println("Request #" + ar.aid + " is timed out and dropped.");
    	  	  	        logFileWriter.println();	
    	  	  	    	dw.requests.removeElementAt(i);
    	  	  	    	i--;
    	  	  	     }    	  	  	        	  	  	
    	  	  }	 	  	   	  	
    	  }    	
    	 
    	logFileWriter.println("Service time on this camera: " + singleServiceTime + " ms.");
 	  	logFileWriter.println("Scheduling time on this camera: " + singleSchedulingTime + " ms."); 	  	 	  	  	  	     	
      } 
      
    public void singleRobotExecution(deviceWorkload dw) 
	  {
		int count = 0;
    	double angle, actionCost;
    	Tuple t;
    	Point loc = new Point();
    	ActionRequest ar;
    	ER1Simulator er1Sim;   	    	
    
    	er1Sim = new ER1Simulator(dw.did);    	
        singleServiceTime = 0; singleSchedulingTime = 0;
        singleSchedulingTime += ER1Simulator.STANDARD_POLLING_TIME;  	    	
    	while (dw.requests.size() != 0)
    	  {
    	  	logFileWriter.println("Current physical status of Robot " + er1Sim.id + ": loc = (" + er1Sim.loc.x + "," + er1Sim.loc.y + "," + er1Sim.loc.z + ") " + "angle = " + er1Sim.angle + ".");
            logFileWriter.println();
                       
            ar = (ActionRequest)dw.requests.elementAt(0);
    	  	t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	actionCost = estimateWarnActionCost(er1Sim.loc,er1Sim.angle,loc);  
    	   	singleServiceTime += actionCost;
    	   	
    	   	angle = computeTargetAngle(er1Sim.loc,loc);
    	   	er1Sim.setLoc(loc.x,loc.y,loc.z);
    	   	er1Sim.setAngle(angle);   	   	
    	   	    	   	
    	  	logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced. ");
    	  	logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	logFileWriter.println();
    	  	    	  	 
    	  	dw.requests.removeElementAt(0); 
    	  	
    	  	for (int i=0;i<dw.requests.size();i++)
    	  	  {    	  	  	
    	  	  	ar = (ActionRequest)dw.requests.elementAt(i);  
    	  	  	
    	  	  	if (ar.deadline != Double.MAX_VALUE)
    	  	  	  if (ar.deadline < overallSchedulingTime + singleSchedulingTime + singleServiceTime)
    	  	  	    {
    	  	  	        logFileWriter.println("Request #" + ar.aid + " is timed out and dropped.");
    	  	  	        logFileWriter.println();	
    	  	  	    	dw.requests.removeElementAt(i);
    	  	  	    	i--;
    	  	  	     }    	  	  	        	  	  	
    	  	  }	 	  	   	  	
    	  }    	
    	 
    	logFileWriter.println("Service time on this robot: " + singleServiceTime + " ms.");
 	  	logFileWriter.println("Scheduling time on this robot: " + singleSchedulingTime + " ms."); 
 	  	
 	  	try {er1Sim.serverSocket.close();}
 	  	catch (Exception e) {e.printStackTrace();}  	 	  	  	  	     	
      }        
	  
	/** Estimate the cost of a {@link Photo} action. */
	public double estimatePhotoActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_POLLING_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      } 
    
    /** Estimate the cost of a {@link Warn} action. */
   public double estimateWarnActionCost (Point currentLoc, double currentAngle, Point targetLoc)
     {
       double actionCost = 0;
       actionCost += Math.sqrt(Math.pow(targetLoc.x-currentLoc.x,2)+Math.pow(targetLoc.y-currentLoc.y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       double deltaAngle = computeTargetAngle(currentLoc, targetLoc);
       deltaAngle -= currentAngle;
	   if (deltaAngle > 180)
	     deltaAngle -= 360;
	   if (deltaAngle <= -180)
	     deltaAngle += 360;
	   actionCost += ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle;
       
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }
     
   public double computeTargetAngle(Point currentLoc, Point targetLoc)
     {
       double currentX = currentLoc.x, currentY = currentLoc.y;
       double targetX = targetLoc.x, targetY = targetLoc.y;  	    
       double deltaAngle; 
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     		     	  	
	      }	
	      
	   return deltaAngle;	
     }	
 } 