/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of an variation of LIST (SJF: Shorest Job First).
 *
 * @see LFM_LFJ
 * @see LIST
 * @see LJF
 * @see AortaSchedulingSimulator#schedule
 */ 
public class SJF {
	
	byte deviceType;
	
	Vector requests;
	
	Vector finishedAID = new Vector (20,10);
	
	PrintWriter logFileWriter;
	
	double overallServiceTime, overallSchedulingTime;
	
	CameraSimulator camSim [];
	ER1Simulator er1Sim [];
	
		
	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 * 
	 * @see ActionRequest
	 */
	public SJF (byte deviceType, Vector requests, PrintWriter logFileWriter) 
	  {
		this.deviceType = deviceType;
		this.requests = requests;
		this.logFileWriter = logFileWriter;	
		
		overallServiceTime = 0;  overallSchedulingTime = 0;	
      }     
   
    public void doAlgorithm () 
      {
    	int i, j,k; 
    	long startTime, endTime;
    	
    	startTime = System.currentTimeMillis();   	
    	  	
    	// decide the set of candidate devices
    	Vector didVector = new Vector (20,10);
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	   	
    	
    	deviceRequestQueue devices [] = new deviceRequestQueue [didVector.size()];
    	if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  camSim = new CameraSimulator [didVector.size()];
    	else
    	  if (deviceType == AortaSchedulingSimulator.ROBOT)
    	    er1Sim = new ER1Simulator [didVector.size()];
    	  else
    	    {
      	      logFileWriter.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      logFileWriter.close();
      	      
      	      System.err.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      System.exit(0);
      	    }
    	for (i=0;i<didVector.size();i++)
    	  {
    	  	if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  	  camSim[i] = new CameraSimulator(String.valueOf(i+1));
    	  	else
    	  	  er1Sim[i] = new ER1Simulator(String.valueOf(i+1));
    	  	    	  	
    	  	devices[i] = new deviceRequestQueue();
    	  	devices[i].did = (String)didVector.elementAt(i);
    	  	for (j=0;j<requests.size();j++)
    	  	  for (k=0;k<((ActionRequest)requests.elementAt(j)).didVector.size();k++)
    	  	    if ( ((String)((ActionRequest)requests.elementAt(j)).didVector.elementAt(k)).equals(devices[i].did) )
    	  	      {
    	  	        devices[i].requests.addElement((ActionRequest)requests.elementAt(j));
    	  	        break;
    	  	      } 	  	
    	  } 
    	  
    	endTime = System.currentTimeMillis();
    	
    	overallSchedulingTime += endTime - startTime;
    	
    	schedule(devices);  		
    }    
    
    /** Schedule the action requests to the candidate devices. */
    public void schedule (deviceRequestQueue devices []) 
      {
    	int i,j,k,m;
    	int count = 0;
    	int finishedNum = 0;
    	int size = requests.size();
    	int id;
    	int pan = 0, tilt = 0, zoom = 1;
    	int deltaPan, deltaTilt, deltaZoom;
    	int pos;
    	long currentTime, startTime, endTime;
    	double pollingTime, headMovementCost, actionCost, angle, minActionCost;        
        double singleServiceTime [] = new double [devices.length];
    	Tuple t;
    	Point loc = new Point();
    	ActionRequest ar;    	    	
    	
    	for (i=0;i<singleServiceTime.length;i++)
    	  singleServiceTime[i] = 0; 
    	   	  
    	if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  {
    	  	pollingTime = 0;
 		    for (i=0;i<camSim.length;i++)
 		      if (pollingTime < camSim[i].polling())
 		        pollingTime = camSim[i].pollingTime; 
 		    overallSchedulingTime += pollingTime;
 		  }
 		else
 		  overallSchedulingTime += ER1Simulator.STANDARD_POLLING_TIME;	
 		
    	logFileWriter.println("#");
    	while (finishedNum != size)
    	  { 
    	    currentTime = System.currentTimeMillis();
    	    
    	  	for (i=0;i<devices.length;i++)
    	  	  if (devices[i].availableTime < currentTime)  // the device is free
    	  	    {
    	  	      startTime = System.currentTimeMillis();
    	  	      for (j=0;j<devices[i].requests.size();j++)
    	  	        if (!finishedAID.contains(((ActionRequest)devices[i].requests.elementAt(j)).aid))
    	  	          break;
    	  	   	      
    	  	      if (j == devices[i].requests.size())  // all requests that this device can service has been serviced
    	  	        {
    	  	          devices[i].availableTime = Double.MAX_VALUE;  // disable the device
    	  	          endTime = System.currentTimeMillis();
    	  	          overallSchedulingTime += endTime - startTime;
    	  	          continue;
    	  	        }
    	  	   	
    	  	   	  id = Integer.parseInt(devices[i].did);
    	  	   	  
    	  	   	  ar = (ActionRequest)devices[i].requests.elementAt(j);
    	  	      for (k=0;k<ar.didVector.size();k++)
    	  	        if (devices[i].did.equals((String)ar.didVector.elementAt(k)))
    	  	          {ar.sel_no = k; break;}    	  	      
    	  	      t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	  	  if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  	        {
    	  	          pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	          tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	          zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	          deltaPan = Math.abs(pan - camSim[id-1].pan);
    	   	          deltaTilt = Math.abs(tilt - camSim[id-1].tilt);
    	   	          deltaZoom = Math.abs(zoom - camSim[id-1].zoom);    	   	    
    	   	          headMovementCost = camSim[id-1].computeHeadMovementCost(deltaPan,deltaTilt, deltaZoom);
    	   	          minActionCost = estimatePhotoActionCost(headMovementCost);
    	   	       
    	  	   	      for (k=j+1;k<devices[i].requests.size();k++)
    	  	   	        if (!finishedAID.contains(((ActionRequest)devices[i].requests.elementAt(k)).aid))
    	  	              {
    	  	          	    ar = (ActionRequest)devices[i].requests.elementAt(k);
    	  	                for (m=0;m<ar.didVector.size();m++)
    	  	                  if (devices[i].did.equals((String)ar.didVector.elementAt(m)))
    	  	                    {ar.sel_no = m; break;}
    	  	                t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	  	            pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	                tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	                zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	                deltaPan = Math.abs(pan - camSim[id-1].pan);
    	   	                deltaTilt = Math.abs(tilt - camSim[id-1].tilt);
    	   	                deltaZoom = Math.abs(zoom - camSim[id-1].zoom);    	   	    
    	   	                headMovementCost = camSim[id-1].computeHeadMovementCost(deltaPan,deltaTilt, deltaZoom);
    	   	                actionCost = estimatePhotoActionCost(headMovementCost);
    	   	                if (minActionCost > actionCost)
    	   	                  {minActionCost = actionCost; j = k;} 
    	  	              }
    	  	        }
    	  	      else
    	  	        {
    	  	          loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	          loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	          loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	          minActionCost = estimateWarnActionCost(er1Sim[id-1].loc,er1Sim[id-1].angle,loc);
    	   	          
    	   	          for (k=j+1;k<devices[i].requests.size();k++)
    	  	   	        if (!finishedAID.contains(((ActionRequest)devices[i].requests.elementAt(k)).aid))
    	  	              {
    	  	          	    ar = (ActionRequest)devices[i].requests.elementAt(k);
    	  	                for (m=0;m<ar.didVector.size();m++)
    	  	                  if (devices[i].did.equals((String)ar.didVector.elementAt(m)))
    	  	                    {ar.sel_no = m; break;}
    	  	                t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	  	            loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	                loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	                loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	                actionCost = estimateWarnActionCost(er1Sim[id-1].loc,er1Sim[id-1].angle,loc);
    	   	                if (minActionCost > actionCost)
    	   	                  {minActionCost = actionCost; j = k;} 
    	  	              }   	   	          	
    	  	        }
    	  	   	  
    	  	   	  ar = (ActionRequest)devices[i].requests.elementAt(j);
    	  	      for (j=0;j<ar.didVector.size();j++)
    	  	        if (devices[i].did.equals((String)ar.didVector.elementAt(j)))
    	  	          {ar.sel_no = j; break;}
    	  	      t = (Tuple)ar.statusVector.elementAt(ar.sel_no);
    	  	  	  if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  	  	    {
    	  	  	      pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	          tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	          zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	          deltaPan = Math.abs(pan - camSim[id-1].pan);
    	   	          deltaTilt = Math.abs(tilt - camSim[id-1].tilt);
    	   	          deltaZoom = Math.abs(zoom - camSim[id-1].zoom);    	   	    
    	   	          headMovementCost = camSim[id-1].computeHeadMovementCost(deltaPan,deltaTilt, deltaZoom);
    	   	          actionCost = estimatePhotoActionCost(headMovementCost); 
    	   	  	  
    	   	  	      camSim[id-1].pan = pan;  	   	  
    	  	   	      camSim[id-1].tilt = tilt;
    	  	   	      camSim[id-1].zoom = zoom;
    	  	   	    }
    	  	   	  else
    	  	   	    {
    	  	   	      loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	          loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	          loc.z = Double.parseDouble((String)t.fields.elementAt(2));
    	   	          actionCost = estimateWarnActionCost(er1Sim[id-1].loc,er1Sim[id-1].angle,loc); 
    	   	          
    	   	          angle = computeTargetAngle(er1Sim[id-1].loc,loc);
    	   	          er1Sim[id-1].setLoc(loc.x,loc.y,loc.z);
    	   	          er1Sim[id-1].setAngle(angle);    		
    	  	   	    }
    	  	   	  
    	  	   	  finishedAID.addElement((String)ar.aid);
    	  	   	  finishedNum ++;
    	  	   	  
    	  	   	  endTime = System.currentTimeMillis();
    	  	   	  overallSchedulingTime += endTime - startTime;
    	  	   	  singleServiceTime[id-1] += actionCost;
    	  	   	  devices[i].availableTime = endTime + actionCost;
    	  	   	  
    	  	   	  logFileWriter.println("Round #" + (++count) + ": Request #" + ar.aid + " is serviced on Device " + id + ".");
    	  	      logFileWriter.println("Action cost: " + actionCost + " ms.");
    	  	      logFileWriter.println();
    	  	      
    	  	      if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  	        logFileWriter.println("Current physical status of Camera " + id + ": pan = " + pan + "  tilt = " + tilt + "  zoom = " + zoom + ".");
    	  	      else
 		  	        logFileWriter.println("Current physical status of Robot " + id + ": loc = (" + er1Sim[id-1].loc.x + "," + er1Sim[id-1].loc.y + "," + er1Sim[id-1].loc.z + ") " + "angle = " + er1Sim[id-1].angle + ".");
                  logFileWriter.println(); 	  	        	  
    	  	    }                  	  	     
    	  } 
    	   
        logFileWriter.println("#");
 	    for (i=0;i<singleServiceTime.length;i++)
 	      {
 	      	logFileWriter.println("Service time on Device " + (i+1) + ": " + singleServiceTime[i] + " ms.");
 	      	if (singleServiceTime[i] > overallServiceTime)
 	      	  overallServiceTime = singleServiceTime[i]; 
 	      	if (i == singleServiceTime.length-1)
 	      	  logFileWriter.println("#");
 	      	logFileWriter.println();	      	
 	      }
 	      
 		try
 	      {
 	      	if (deviceType == AortaSchedulingSimulator.ROBOT)
 	          for (i=0;i<er1Sim.length;i++)
 	            er1Sim[i].serverSocket.close();
 	       }
    	catch (Exception e) {e.printStackTrace();}
 		 		           
        logFileWriter.println("#");
 		logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 		logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 		logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 		logFileWriter.println("#");  	
      }      
    
    /** Estimate the cost of a {@link Photo} action. */
    public double estimatePhotoActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_POLLING_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      } 
      
    /** Estimate the cost of a {@link Warn} action. */
   public double estimateWarnActionCost (Point currentLoc, double currentAngle, Point targetLoc)
     {
       double actionCost = 0;
       actionCost += Math.sqrt(Math.pow(targetLoc.x-currentLoc.x,2)+Math.pow(targetLoc.y-currentLoc.y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       double deltaAngle = computeTargetAngle(currentLoc, targetLoc);
       deltaAngle -= currentAngle;
	   if (deltaAngle > 180)
	     deltaAngle -= 360;
	   if (deltaAngle <= -180)
	     deltaAngle += 360;
	   actionCost += ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle;
       
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }
     
   public double computeTargetAngle(Point currentLoc, Point targetLoc)
     {
       double currentX = currentLoc.x, currentY = currentLoc.y;
       double targetX = targetLoc.x, targetY = targetLoc.y;  	    
       double deltaAngle; 
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     		     	  	
	      }	
	      
	   return deltaAngle;	
     }      		
 } 