/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of Algorithm 2 (SRFAE: Shorest Request First Assignment and Execution).
 * 
 * @see LRFAE
 * @see AortaSchedulingSimulator#schedule
 * @see SinglePassThread#scheduleCameras
 * @see SinglePassThread#scheduleRobots
 */ 
public class SRFAE {
 	
 	byte deviceType;
 	
 	Vector requests;
	
	PrintWriter logFileWriter;
	
	Vector didVector = new Vector (20,10);
	
	CameraSimulator camSim [];
	ER1Simulator er1Sim [];
	double singleServiceTime [];
	double overallServiceTime, overallSchedulingTime;
	
	ActionRequest sel_ar;
	String sel_did;
	Tuple sel_status;
	double total_actionCost, sel_actionCost;	
	
		
 	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 * 
	 * @see ActionRequest
	 */
	 public SRFAE (byte deviceType, Vector requests, PrintWriter logFileWriter) 
 	  {
 		this.deviceType = deviceType;
 		this.requests = requests;
		this.logFileWriter = logFileWriter;	
		
		overallServiceTime = 0;  overallSchedulingTime = 0; 
		
		long startTime = System.currentTimeMillis(); 				
 		 			
 		initialize();
 		
 		long endTime = System.currentTimeMillis();
 		
 		overallSchedulingTime += endTime - startTime;		  
 	  }  	
 	
 	public void initialize ()
 	  {
 	  	// decide the set of candidate devices
		int i,j;
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	
    	if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  camSim = new CameraSimulator [didVector.size()];
    	else
    	  if (deviceType == AortaSchedulingSimulator.ROBOT)
    	    er1Sim = new ER1Simulator [didVector.size()];
    	  else
    	    {
      	      logFileWriter.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      logFileWriter.close();
      	      
      	      System.err.println("Unsupported device type. Must be \"camera\" or \"robot\".");
      	      System.exit(0);
      	    }
    	singleServiceTime = new double [didVector.size()];    	
       	for (i=0;i<didVector.size();i++)
    	  {
    	  	if (deviceType == AortaSchedulingSimulator.CAMERA)
    	  	  camSim[i] = new CameraSimulator(String.valueOf(i+1));
    	  	else 
    	  	  er1Sim[i] = new ER1Simulator(String.valueOf(i+1));
    	  	  
    	  	singleServiceTime[i] = 0;    	  	
    	  }  	  	
 	  } 	  
 	  
 	/** Assign and execute the requests on the candidate devices with the shorest request first heurisic. */
 	public void execution ()  
 	  {
 		int i;
 		int count = 0;
 		int id;
 		long startTime, endTime;
 		double pollingTime, angle, actionCost;
 		Point loc = new Point(); 	
 		
 		if (deviceType == AortaSchedulingSimulator.CAMERA)
 		  {
 		  	pollingTime = 0;
 		    for (i=0;i<camSim.length;i++)
 		      if (pollingTime < camSim[i].polling())
 		        pollingTime = camSim[i].pollingTime; 
 		    overallSchedulingTime += pollingTime;
 		  }
 		else
 		  overallSchedulingTime += ER1Simulator.STANDARD_POLLING_TIME; 		  	
 		
 		while (requests.size() > 0)
 		  {
 		  	logFileWriter.println("#"); 		  	
             		  	
 		  	startTime = System.currentTimeMillis(); 		  	
 		  	shorestRequestSelection();
 		  	endTime = System.currentTimeMillis();
 		  	overallSchedulingTime += endTime - startTime;
 		  	
 		  	if (total_actionCost == Double.MAX_VALUE)
 		  	  {
 		  	  	logFileWriter.println("#");
 		  	  	
 		  	  	for (i=0;i<requests.size();i++)
 		  	  	  {
 		  	  	  	logFileWriter.println("Request #" + ((ActionRequest)requests.elementAt(i)).aid + " is timed out and dropped.");
 		  	  	  	logFileWriter.println();
 		  	  	  }
 		  	  	break;
 		  	  }
 		  	  	
 		  	id = Integer.parseInt(sel_did);
 		  	sel_actionCost = total_actionCost - singleServiceTime[id-1];
 		  	singleServiceTime[id-1] = total_actionCost;  		  	 		  	
 		  	
 		  	logFileWriter.print("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is serviced on ");
 		  	if (deviceType == AortaSchedulingSimulator.CAMERA)
 		  	  logFileWriter.print("Camera ");
 		  	else
 		  	  logFileWriter.print("Robot "); 		  	  
 		  	logFileWriter.println(sel_did + ".");
 		  	logFileWriter.println("Action cost: " + sel_actionCost + " ms.");
 		  	logFileWriter.println("#");
 		  	logFileWriter.println();
 		  	 		  	
 		  	if (deviceType == AortaSchedulingSimulator.CAMERA)
 		  	  {
 		  	  	camSim[id-1].pan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	    camSim[id-1].tilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	    camSim[id-1].zoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	  }
 		  	else
 		  	  {
 		  	  	loc.x = Double.parseDouble((String)sel_status.fields.elementAt(0));
    	   	    loc.y = Double.parseDouble((String)sel_status.fields.elementAt(1));
    	   	    loc.z = Double.parseDouble((String)sel_status.fields.elementAt(2));
    	   	    angle = computeTargetAngle(er1Sim[id-1].loc,loc);
    	   	    er1Sim[id-1].setLoc(loc.x,loc.y,loc.z);
    	   	    er1Sim[id-1].setAngle(angle); 
 		  	  }
 		  	
 		  	if (deviceType == AortaSchedulingSimulator.CAMERA)
 		  	  logFileWriter.println("Current physical status of Camera " + id + ": pan = " + camSim[id-1].pan + "  tilt = " + camSim[id-1].tilt + "  zoom = " + camSim[id-1].zoom + ".");
 		  	else
 		  	  logFileWriter.println("Current physical status of Robot " + id + ": loc = (" + er1Sim[id-1].loc.x + "," + er1Sim[id-1].loc.y + "," + er1Sim[id-1].loc.z + ") " + "angle = " + er1Sim[id-1].angle + ".");
            logFileWriter.println();
            		  	
 		  	startTime = System.currentTimeMillis();
 		  	requests.removeElement((ActionRequest)sel_ar); 		  	
 		  	if (deviceType == AortaSchedulingSimulator.CAMERA)
 		  	  updatePhotoCosts();
 		  	else
 		  	  updateWarnCosts();
 		  	endTime = System.currentTimeMillis();
 		  	overallSchedulingTime += endTime - startTime;		  		     	
 		  } 
 	 		  
 	    logFileWriter.println("#");
 	    for (i=0;i<didVector.size();i++)
 	      {
 	      	logFileWriter.println("Service time on Device " + (i+1) + ": " + singleServiceTime[i] + " ms.");
 	      	if (singleServiceTime[i] > overallServiceTime)
 	      	  overallServiceTime = singleServiceTime[i]; 
 	      	if (i == singleServiceTime.length-1)
 	      	  logFileWriter.println("#");
 	      	logFileWriter.println();		      	
 	      }
 	      
 	    try
 	      {
 	      	if (deviceType == AortaSchedulingSimulator.ROBOT)
 	          for (i=0;i<er1Sim.length;i++)
 	            er1Sim[i].serverSocket.close();
 	       }
    	catch (Exception e) {e.printStackTrace();}
 		
 		logFileWriter.println("#");
 		logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 		logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 		logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 		logFileWriter.println("#");
 	  }  	
 	
 	/** Select the current shorest request on all canidate devices. */
 	public void shorestRequestSelection() 
 	  {
		int i,j;
		double minWeight = Double.MAX_VALUE, minActionCost=Double.MAX_VALUE;
		double weight;
		ActionRequest ar;
		
		total_actionCost = minActionCost;
		
		for (i=0;i<requests.size();i++)
		  {
		    ar = (ActionRequest)requests.elementAt(i);	
		    for (j=0;j<ar.costVector.size();j++)
		      {
		      	if (ar.deadline == Double.MAX_VALUE)
		      	  weight = Double.parseDouble((String)ar.costVector.elementAt(j));
		      	else
		      	  if (ar.deadline < (overallSchedulingTime + singleServiceTime[Integer.parseInt((String)ar.didVector.elementAt(j))-1]) )
		      	    continue;
		      	  else
		      	    {
		      	      weight = UniformTCSyntheticWorkload.alpha*Double.parseDouble((String)ar.costVector.elementAt(j)) + (1-UniformTCSyntheticWorkload.alpha)*ar.deadline;
		      	      // System.out.println(weight);
		      	    }
		         
		        if (minWeight > weight)
		          {
		      	    sel_ar = ar;
		      	    minActionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));
		      	    sel_did = (String)sel_ar.didVector.elementAt(j);
		      	    sel_status = (Tuple)sel_ar.statusVector.elementAt(j);
		      	    total_actionCost = minActionCost;
		      	    minWeight = weight;			      		      	
		          }
		      }
		  }			 
      }	
	
	/** Update the cost of other requests on the camera that services the currently selected {@link Photo} action request. */
	public void updatePhotoCosts() 
	  {
		int i, j;
		int pan, tilt, zoom;
		int deltaPan, deltaTilt, deltaZoom;
		double headMovementCost, actionCost;
		Tuple t;
		ActionRequest ar;		
		
		for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	for (j=0;j<ar.didVector.size();j++)
		      if (((String)ar.didVector.elementAt(j)).equals(sel_did))
		       {
		         t = (Tuple)ar.statusVector.elementAt(j);
		         pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	     tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	     zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	     deltaPan = Math.abs(camSim[Integer.parseInt(sel_did)-1].pan - pan);
    	   	     deltaTilt = Math.abs(camSim[Integer.parseInt(sel_did)-1].tilt - tilt);
    	   	     deltaZoom = Math.abs(camSim[Integer.parseInt(sel_did)-1].zoom - zoom);
    	   	     headMovementCost = camSim[Integer.parseInt(sel_did)-1].computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
		         actionCost = estimatePhotoActionCost(headMovementCost);     	   	  	
		         actionCost += total_actionCost;
		         
		         ar.costVector.removeElementAt(j);
		         ar.costVector.insertElementAt(String.valueOf(actionCost),j);
		       } 
		   }	
	  }	
	  
	/** Update the cost of other requests on the robot that services the currently selected {@link Warn} action request. */
	public void updateWarnCosts()
	  {
	  	int i, j;	  	
		double actionCost;
		Tuple t;
		Point loc = new Point();
		ActionRequest ar;		
		
		for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	for (j=0;j<ar.didVector.size();j++)
		      if (((String)ar.didVector.elementAt(j)).equals(sel_did))
		       {
		         t = (Tuple)ar.statusVector.elementAt(j);
		         loc.x = Double.parseDouble((String)t.fields.elementAt(0));
    	   	     loc.y = Double.parseDouble((String)t.fields.elementAt(1));
    	   	     loc.z = Double.parseDouble((String)t.fields.elementAt(2));    	   	     
    	   	     actionCost = estimateWarnActionCost(er1Sim[Integer.parseInt(sel_did)-1].loc,er1Sim[Integer.parseInt(sel_did)-1].angle,loc);      	   	  	
		         actionCost += total_actionCost;
		         
		         ar.costVector.removeElementAt(j);
		         ar.costVector.insertElementAt(String.valueOf(actionCost),j);
		       } 
		   }
	  }  
	  
	/** Estimate the cost of a {@link Photo} action. */
	public double estimatePhotoActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_POLLING_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      }
      
    /** Estimate the cost of a {@link Warn} action. */
   public double estimateWarnActionCost (Point currentLoc, double currentAngle, Point targetLoc)
     {
       double actionCost = 0;
       actionCost += Math.sqrt(Math.pow(targetLoc.x-currentLoc.x,2)+Math.pow(targetLoc.y-currentLoc.y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       
       double deltaAngle = computeTargetAngle(currentLoc, targetLoc);
       deltaAngle -= currentAngle;
	   if (deltaAngle > 180)
	     deltaAngle -= 360;
	   if (deltaAngle <= -180)
	     deltaAngle += 360;
	   actionCost += ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle;
       
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }      
      
    public double computeTargetAngle(Point currentLoc, Point targetLoc)
     {
       double currentX = currentLoc.x, currentY = currentLoc.y;
       double targetX = targetLoc.x, targetY = targetLoc.y;  	    
       double deltaAngle; 
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }     		     	  	
	      }	
	      
	   return deltaAngle;	
     }
 } 