/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;
import java.net.*;

/**
 * Thread for a single pass of scheduling in a shared action operator.
 * 
 * @see GroupingThread
 * @see ActionOperator 
 */ 
public class SinglePassThread extends Thread {
	
	ActionOperator actionOp;
	
	/** A {@link PrintWriter} corresponds to the log file. **/
    PrintWriter logFileWriter;
    
    int id;
    
    long opStartTime, opEndTime;
    
	
	SinglePassThread (ActionOperator actionOp, int id)
	  {
	  	this.actionOp = actionOp;
	  	
	  	this.id = id;
	  }	  
	
	/**
     * Perform a round of scheduling.
     *
     * @see ActionRequest 
     */
	public void run ()
	  {
	  	opStartTime = System.currentTimeMillis();
	  	
	  	String logFile = "../logs/scheduling_logs/" + actionOp.action.name + "_" + String.valueOf(id) + ".log";
	   	
      	try {logFileWriter = new PrintWriter ( new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		    
		logFileWriter.println("A grouping time interval begins.");
	   	System.out.println();
	   	System.out.println("A grouping time interval begins."); 	   	        
			  	    
	  	if ((actionOp.q.selOp != null) && (actionOp.q.selOp.outputOperator == actionOp))
	  	  {
	  	   	actionOp.q.selOp.getInput();
	  	      		  	      	
	  	   	actionOp.inputQueue.removeAllElements();  // remove the event tuple pushed from the scan operator
	  	      	
	  	   	if (actionOp.inputQueue2.size() == 0)  // no candidate device
	  	   	  return;
	  	     	  	  	      	
	  	   	schedule(actionOp.inputQueue2);	  	      	
	      }
	    else
	  	   if ((actionOp.q.joinOp != null) && (actionOp.q.joinOp.outputOperator == actionOp))
	  	     {
	  	       if (actionOp.inputQueue.size() == 0)
	  	         return;
	  	            
	  	       schedule(actionOp.inputQueue);
	  	     }
	  	    
	  	 logFileWriter.println("A grouping time interval ends");
	  	 System.out.println("A grouping time interval ends.");
	  	 System.out.print("Aorta>");
	  	 logFileWriter.println("Optimization time (not including the scheduling time of the scheduling algorithm): " + (opEndTime - opStartTime) + " ms.");
	  	 logFileWriter.println();
	  	 logFileWriter.flush();
	  	 logFileWriter.close();
	  }
	  
	/** 
     * Schedule the requests based on the type of device that the action involves. 
     *
     * @param v a vector contains the candidate tuples of the requests
     * 
     * @see #scheduleSensors
     * @see #scheduleCameras
     * @see #schedulePhones
     * @see #scheduleRobots
     */
    public void schedule (Vector v)
	  {
	  	if (v.size() == 0)
	  	  return;
	  	
	  	if (actionOp.action.device.type.equals("sensor"))
	  	  scheduleSensors(v);
	  	else
	  	  if (actionOp.action.device.type.equals("camera"))
	  	    scheduleCameras(v);
	  	  else
	  	    if (actionOp.action.device.type.equals("phone"))
	  	      schedulePhones(v);
	  	    else
	  	      if (actionOp.action.device.type.equals("robot"))
	  	        scheduleRobots(v);        
	  }	  
	  
    public void scheduleSensors (Vector v)
      {
      	// not yet implemented
      }      
    
    /**
     * @see LFRFA
     * @see SRFE
     * @see SRFAE
     * @see SA
     * @see LFM_LFJ
     * @see RANDOM
     * @see AortaSchedulingSimulator  
     */
    public void scheduleCameras (Vector v)
      {
      	int i,j;
      	int count;
      	int mark;
      	int id, qid;
      	int pan, tilt, zoom;
      	long seqNo;
      	long startTime, endTime;
      	double estimatedCost;
      	String camera_id;
      	Vector didVector = new Vector(10,5);
      	Vector dparamsVector = new Vector (10,5);
      	Vector requests = new Vector (20, 10);
      	Tuple t;
      	ActionRequest ar;
      	CameraPollingThread cpt [];
      		
      	  	
      	// separate different action requests
      	logFileWriter.println("#");
      	
      	count = 1;
      	Vector vv = new Vector (10,5);
      	while (v.size() != 0)
      	  {
      	  	Vector vt = new Vector (10,5);
      	  	boolean locked = true;
      	  	
      	  	ar = new ActionRequest();
      	  	ar.aid = String.valueOf(count);
      	
         	t = (Tuple)v.elementAt(0);
      	    qid = t.qid;
      	    seqNo = t.seqNo;
      	    camera_id = (String)t.fields.elementAt(0);
      	    if (!actionOp.q.aortaqp.lockCameraVector.contains(camera_id)) 
      	      locked = false;
      	    if (!didVector.contains(camera_id))
      	      {
      	      	didVector.addElement((String)camera_id);
      	      	dparamsVector.addElement((Tuple)t);
      	      }
      	    ar.didVector.addElement((String)camera_id);
      	    ar.statusVector.addElement((Tuple)t);      	    
      	    vt.addElement((Tuple)v.elementAt(0));
      	    v.removeElementAt(0);
      	    
      	    logFileWriter.println("Action Request #" + (count++) + ":");
      	    logFileWriter.println("Candidate camera: Camera " + camera_id);
      	     
      	    while (true)
      	      {
      	      	if (v.size() == 0) 
      	      	  break;
      	      	t = (Tuple)v.elementAt(0);
      	      	if (t.qid == qid && t.seqNo == seqNo)
      	      	  {
      	      	  	camera_id = (String)t.fields.elementAt(0);
      	      	  	logFileWriter.println("Candidate camera: Camera " + camera_id);
      	      	  	if (!actionOp.q.aortaqp.lockCameraVector.contains(camera_id)) 
      	              locked = false;
      	            if (!didVector.contains(camera_id))
      	              {
      	              	didVector.addElement((String)camera_id);
      	              	dparamsVector.addElement((Tuple)t);
      	              } 
      	            ar.didVector.addElement((String)camera_id);
      	            ar.statusVector.addElement((Tuple)t);      	            
      	            vt.addElement((Tuple)v.elementAt(0));
      	            v.removeElementAt(0);
      	          }
      	      	else
      	      	  break;      	      	
      	      } 
      	      
      	    if (locked == true)
      	      {
      	      	for (i=0;i<vt.size();i++)
      	          vv.addElement((Tuple)v.elementAt(i));
      	        count --;
      	      }
      	    else
      	      requests.addElement((ActionRequest)ar); 
      	    logFileWriter.println();	           	    
      	  }      	  
      	
      	v = vv;
      	
      	// poll the candidate devices
      	logFileWriter.println("#");
	   	logFileWriter.println("Polling the candidate cameras .... ");
	   	
	   	cpt = new CameraPollingThread[didVector.size()];
	   	
	   	startTime = System.currentTimeMillis();
	   	
	   	for (i=0;i<didVector.size();i++)
	   	  {
	   	    camera_id = (String)didVector.elementAt(i);	   	      	
	   	      	
	   	    if (actionOp.camSim[Integer.parseInt(camera_id)-1] == null)
	   	      actionOp.camSim[Integer.parseInt(camera_id)-1] = new CameraSimulator(camera_id);
	   	      	  
	   	    if (actionOp.q.aortaqp.lockingEnabled)
	   	      if (actionOp.q.aortaqp.lockCameraVector.contains(camera_id))
	   	        {cpt[i] = null; continue;} 
	   	    
	   	    t = (Tuple)dparamsVector.elementAt(i);      
	   	    cpt[i] = new CameraPollingThread(actionOp.camSim[Integer.parseInt(camera_id)-1], t, actionOp.q.aortaqp.CAMERA_SIMULATION);
	   	    cpt[i].start();
	      }	
	         	    
	   	endTime = System.currentTimeMillis();
	   		   	        
	   	while (endTime - startTime < CameraSimulator.POLLING_TIMEOUT_LIMIT)
	   	  {
	   	    boolean finished = true;
	   	          	
	   	    for (i=0;i<didVector.size();i++)
	   	      if (cpt[i] != null && cpt[i].connected == false)
	   	        finished = false;
	   	          	    	   	          	
	   	    if (finished) break;
	   	       
	   	    endTime = System.currentTimeMillis();	   	          	
	   	  } 
	   	   	
      	
      	// remove those candidate devices the polling to which timed out
      	for (i=0;i<didVector.size();i++)
	   	  if (cpt[i] == null || cpt[i].connected == false)
	   	      {didVector.removeElementAt(i); i--;}	   	
	   	
	   	for (i=0;i<requests.size();i++)
	   	  {
	   	  	ar = (ActionRequest)requests.elementAt(i);
	   	  	for (j=0;j<ar.didVector.size();j++)
	   	  	  {
	   	  	  	camera_id = (String)ar.didVector.elementAt(j);
	   	  	  	if (!didVector.contains((String)camera_id))
	   	  	  	  {
	   	  	  	  	ar.didVector.removeElementAt(j);
	   	  	  	  	ar.statusVector.removeElementAt(j);
	   	  	  	  	j--;
	   	  	  	  }
	   	  	  }
	   	  }
	   	  
	   	logFileWriter.println("Finish polling the candidate cameras.");
	   	logFileWriter.println("#");
	   	logFileWriter.flush();	   	   
	   	
	   	// compute the estimated cost of each action request on each candidate device of it    
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	t = (Tuple)ar.statusVector.elementAt(j);      	  	  	      	  	  	  	  	  	
	   	        actionOp.camera_id_c = (String)t.fields.elementAt(0);  	        
	   	        actionOp.loc_c = Point.parsePoint((String)t.fields.elementAt(4));
	   	        
	   	        estimatedCost = actionOp.estimateActionCost();
	   	        logFileWriter.println("Estimated action cost on candidate Camera " + actionOp.camera_id_c + " of Request #" + ar.aid + " is:" + estimatedCost + " ms.");   	        	   	        
	   	        ar.costVector.addElement(String.valueOf(estimatedCost));
	   	        
	   	        id = Integer.parseInt(actionOp.camera_id_c);
	   	        pan = actionOp.camSim[id-1].targetPan(actionOp.loc_c);
	   	        tilt = actionOp.camSim[id-1].targetTilt(actionOp.loc_c);
	   	        zoom = actionOp.camSim[id-1].targetZoom(actionOp.loc_c);
	   	        
	   	        Vector vt = new Vector (10,5);
	   	        vt.addElement(String.valueOf(pan));
	   	        vt.addElement(String.valueOf(tilt));
	   	        vt.addElement(String.valueOf(zoom));
	   	        t = new Tuple(vt);	   	        
	   	        ar.statusVector.removeElementAt(j);
	   	        ar.statusVector.insertElementAt((Tuple)t,j);
	   	      }
	   	    
	   	    logFileWriter.println();    	     	  	
      	  }
      	logFileWriter.println("#");
      	logFileWriter.flush();
      	
      	// lock the available candidate cameras
      	if (actionOp.q.aortaqp.lockingEnabled)
      	  for (i=0;i<didVector.size();i++)
      	    actionOp.q.aortaqp.lockCameraVector.addElement((String)didVector.elementAt(i));
      	      	
      	
      	opEndTime = System.currentTimeMillis(); 
      	
      	
      	// schedule the requests
      	switch (actionOp.q.aortaqp.schedulingAlgorithm)
      	  {
      	  	case Aorta.LFRFA_PLUS_SRFE:
      	  	  {
      	  	  	logFileWriter.println("Scheduling algorithm: LFRFA + SRFE. ");
		        System.out.println("Scheduling algorithm: LFRFA + SRFE. ");
		        
		        startTime = System.currentTimeMillis();
	            LFRFA lfrfa = new LFRFA (requests);
		        lfrfa.assignment();
		        endTime = System.currentTimeMillis();
		        logFileWriter.println();		 
		        logFileWriter.println("LFRFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		        SRFE srfe = new SRFE (AortaSchedulingSimulator.CAMERA, requests, logFileWriter, endTime - startTime);
	         	srfe.execution();
      	  	  }
      	  	break;
      	  	
      	  	case Aorta.SRFAE:
      	  	  {
      	  	  	logFileWriter.println("Scheduling algorithm: SRFAE. ");
      	  	  	System.out.println("Scheduling algorithm: SRFAE. ");
		        
		        SRFAE srfae = new SRFAE (AortaSchedulingSimulator.CAMERA, requests, logFileWriter);
		        srfae.execution();		
      	  	  }
      	  	break;
      	  
      	    case Aorta.LFM_LFJ:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: LIST. ");
      	      	System.out.println("Scheduling algorithm: LIST. ");
      	      	
      	      	LFM_LFJ lfm_lfj = new LFM_LFJ (AortaSchedulingSimulator.CAMERA, requests, logFileWriter);
		        lfm_lfj.doAlgorithm();
      	      }
      	    break;
      	    
      	    case Aorta.SA:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: SA. ");
      	      	System.out.println("Scheduling algorithm: SA. ");
      	      	
      	      	SA sa = new SA (AortaSchedulingSimulator.CAMERA, requests, logFileWriter);
		        sa.execution();
      	      }
      	    break;
      	    
      	    case Aorta.RANDOM:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: RANDOM. ");
      	      	System.out.println("Scheduling algorithm: RANDOM. ");
      	      	
      	      	RANDOM random = new RANDOM (AortaSchedulingSimulator.CAMERA, requests, logFileWriter);
		        random.execution();
      	      }      	  	
      	  }
      	logFileWriter.println("#");
      	
      	
      	// unlock the available candidate cameras
      	if (actionOp.q.aortaqp.lockingEnabled)
      	  for (i=0;i<didVector.size();i++)
      	    actionOp.q.aortaqp.lockCameraVector.removeElement((String)didVector.elementAt(i));      	    	     	
      }
      
    public void schedulePhones (Vector v)
      {
      	// not yet implemented
      }
      
    /**
     * @see LFRFA
     * @see SRFE
     * @see SRFAE
     * @see SA
     * @see LFM_LFJ
     * @see RANDOM
     * @see AortaSchedulingSimulator  
     */
    public void scheduleRobots (Vector v)
      {
      	int i,j;
      	int count;
      	int mark;
      	int id, qid;      	
      	long seqNo;
      	long startTime, endTime;
      	double angle, estimatedCost;
      	String robot_id;
      	Vector didVector = new Vector(10,5);
      	Vector dparamsVector = new Vector (10,5);
      	Vector requests = new Vector (20, 10);
      	Tuple t;
      	Point loc = new Point();
      	ActionRequest ar;      	
      	RobotPollingThread rpt [];
      		
      	  	
      	// separate different action requests
      	logFileWriter.println("#");
      	
      	count = 1;
      	Vector vv = new Vector (10,5);
      	while (v.size() != 0)
      	  {
      	  	Vector vt = new Vector (10,5);
      	  	boolean locked = true;
      	  	
      	  	ar = new ActionRequest();
      	  	ar.aid = String.valueOf(count);
      	
         	t = (Tuple)v.elementAt(0);
      	    qid = t.qid;
      	    seqNo = t.seqNo;
      	    robot_id = (String)t.fields.elementAt(0);
      	    if (!actionOp.q.aortaqp.lockRobotVector.contains(robot_id)) 
      	      locked = false;
      	    if (!didVector.contains(robot_id))
      	      {
      	      	didVector.addElement((String)robot_id);
      	      	dparamsVector.addElement((Tuple)t);
      	      }
      	    ar.didVector.addElement((String)robot_id);
      	    ar.statusVector.addElement((Tuple)t);      	    
      	    vt.addElement((Tuple)v.elementAt(0));
      	    v.removeElementAt(0);
      	    
      	    logFileWriter.println("Action Request #" + (count++) + ":");
      	    logFileWriter.println("Candidate robot: Robot " + robot_id);
      	     
      	    while (true)
      	      {
      	      	if (v.size() == 0) 
      	      	  break;
      	      	t = (Tuple)v.elementAt(0);
      	      	if (t.qid == qid && t.seqNo == seqNo)
      	      	  {
      	      	  	robot_id = (String)t.fields.elementAt(0);
      	      	  	logFileWriter.println("Candidate robot: Robot " + robot_id);
      	      	  	if (!actionOp.q.aortaqp.lockCameraVector.contains(robot_id)) 
      	              locked = false;
      	            if (!didVector.contains(robot_id))
      	              {
      	              	didVector.addElement((String)robot_id);
      	              	dparamsVector.addElement((Tuple)t);
      	              } 
      	            ar.didVector.addElement((String)robot_id);
      	            ar.statusVector.addElement((Tuple)t);      	            
      	            vt.addElement((Tuple)v.elementAt(0));
      	            v.removeElementAt(0);
      	          }
      	      	else
      	      	  break;      	      	
      	      } 
      	      
      	    if (locked == true)
      	      {
      	      	for (i=0;i<vt.size();i++)
      	          vv.addElement((Tuple)v.elementAt(i));
      	        count --;
      	      }
      	    else
      	      requests.addElement((ActionRequest)ar); 
      	    logFileWriter.println();	           	    
      	  }      	  
      	
      	v = vv;
      	
      	// poll the candidate devices
      	logFileWriter.println("#");
	   	logFileWriter.println("Polling the candidate robots .... ");
	   	
	   	rpt = new RobotPollingThread[didVector.size()];
	   	
	   	startTime = System.currentTimeMillis();
	   	
	   	for (i=0;i<didVector.size();i++)
	   	  {
	   	    robot_id = (String)didVector.elementAt(i);	   	      	
	   	      	
	   	    if (actionOp.er1Sim[Integer.parseInt(robot_id)-1] == null)
	   	      actionOp.er1Sim[Integer.parseInt(robot_id)-1] = new ER1Simulator(robot_id);
	   	      	  
	   	    if (actionOp.q.aortaqp.lockingEnabled)
	   	      if (actionOp.q.aortaqp.lockRobotVector.contains(robot_id))
	   	        {rpt[i] = null; continue;} 
	   	    
	   	    t = (Tuple)dparamsVector.elementAt(i);      
	   	    rpt[i] = new RobotPollingThread(actionOp.er1Sim[Integer.parseInt(robot_id)-1], (String)t.fields.elementAt(2), actionOp.q.aortaqp);
	   	    rpt[i].start();
	      }	
	         	    
	   	endTime = System.currentTimeMillis();
	   		   	        
	   	while (endTime - startTime < CameraSimulator.POLLING_TIMEOUT_LIMIT)
	   	  {
	   	    boolean finished = true;
	   	          	
	   	    for (i=0;i<didVector.size();i++)
	   	      if (rpt[i] != null && rpt[i].connected == false)
	   	        finished = false;
	   	          	    	   	          	
	   	    if (finished) break;
	   	       
	   	    endTime = System.currentTimeMillis();	   	          	
	   	  } 
	   	   	
      	
      	// remove those candidate devices the polling to which timed out
      	for (i=0;i<didVector.size();i++)
	   	  if (rpt[i] == null || rpt[i].connected == false)
	   	      {didVector.removeElementAt(i); i--;}	   	
	   	
	   	for (i=0;i<requests.size();i++)
	   	  {
	   	  	ar = (ActionRequest)requests.elementAt(i);
	   	  	for (j=0;j<ar.didVector.size();j++)
	   	  	  {
	   	  	  	robot_id = (String)ar.didVector.elementAt(j);
	   	  	  	if (!didVector.contains((String)robot_id))
	   	  	  	  {
	   	  	  	  	ar.didVector.removeElementAt(j);
	   	  	  	  	ar.statusVector.removeElementAt(j);
	   	  	  	  	j--;
	   	  	  	  }
	   	  	  }
	   	  }
	   	  
	   	logFileWriter.println("Finish polling the candidate robots.");
	   	logFileWriter.println("#");
	   	logFileWriter.flush();	   	   
	   	
	   	// compute the estimated cost of each action request on each candidate device of it    
      	for (i=0;i<requests.size();i++)
      	  {
      	  	ar = (ActionRequest)requests.elementAt(i);
      	  	
      	  	for (j=0;j<ar.didVector.size();j++)
      	  	  {
      	  	  	t = (Tuple)ar.statusVector.elementAt(j);      	  	  	      	  	  	  	  	  	
	   	        actionOp.robot_id_c = (String)t.fields.elementAt(0);  	        
	   	        actionOp.loc_c = Point.parsePoint((String)t.fields.elementAt(1));
	   	        
	   	        estimatedCost = actionOp.estimateActionCost();
	   	        logFileWriter.println("Estimated action cost on candidate Robot " + actionOp.robot_id_c + " of Request #" + ar.aid + " is:" + estimatedCost + " ms.");   	        	   	        
	   	        ar.costVector.addElement(String.valueOf(estimatedCost));
	   	        
	   	        Vector vt = new Vector (10,5);
	   	        vt.addElement(String.valueOf(actionOp.loc_c.x));
	   	        vt.addElement(String.valueOf(actionOp.loc_c.y));
	   	        vt.addElement(String.valueOf(actionOp.loc_c.z));
	   	        t = new Tuple(vt);	   	        
	   	        ar.statusVector.removeElementAt(j);
	   	        ar.statusVector.insertElementAt((Tuple)t,j);
	   	      }
	   	    
	   	    logFileWriter.println();    	     	  	
      	  }
      	logFileWriter.println("#");
      	logFileWriter.flush();
      	
      	// lock the available candidate robots
      	if (actionOp.q.aortaqp.lockingEnabled)
      	  for (i=0;i<didVector.size();i++)
      	    actionOp.q.aortaqp.lockRobotVector.addElement((String)didVector.elementAt(i));
      	      	
      	
      	opEndTime = System.currentTimeMillis(); 
      	
      	
      	// schedule the requests
      	switch (actionOp.q.aortaqp.schedulingAlgorithm)
      	  {
      	  	case Aorta.LFRFA_PLUS_SRFE:
      	  	  {
      	  	  	logFileWriter.println("Scheduling algorithm: LFRFA + SRFE. ");
		        System.out.println("Scheduling algorithm: LFRFA + SRFE. ");
		        
		        startTime = System.currentTimeMillis();
	            LFRFA lfrfa = new LFRFA (requests);
		        lfrfa.assignment();
		        endTime = System.currentTimeMillis();
		        logFileWriter.println();		 
		        logFileWriter.println("LFRFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		        SRFE srfe = new SRFE (AortaSchedulingSimulator.ROBOT, requests, logFileWriter, endTime - startTime);
	         	srfe.execution();
      	  	  }
      	  	break;
      	  	
      	  	case Aorta.SRFAE:
      	  	  {
      	  	  	logFileWriter.println("Scheduling algorithm: SRFAE. ");
      	  	  	System.out.println("Scheduling algorithm: SRFAE. ");
		        
		        SRFAE srfae = new SRFAE (AortaSchedulingSimulator.ROBOT, requests, logFileWriter);
		        srfae.execution();		
      	  	  }
      	  	break;
      	  
      	    case Aorta.LFM_LFJ:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: LIST. ");
      	      	System.out.println("Scheduling algorithm: LIST. ");
      	      	
      	      	LFM_LFJ lfm_lfj = new LFM_LFJ (AortaSchedulingSimulator.ROBOT, requests, logFileWriter);
		        lfm_lfj.doAlgorithm();
      	      }
      	    break;
      	    
      	    case Aorta.SA:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: SA. ");
      	      	System.out.println("Scheduling algorithm: SA. ");
      	      	
      	      	SA sa = new SA (AortaSchedulingSimulator.ROBOT, requests, logFileWriter);
		        sa.execution();
      	      }
      	    break;
      	    
      	    case Aorta.RANDOM:
      	      {
      	      	logFileWriter.println("Scheduling algorithm: RANDOM. ");
      	      	System.out.println("Scheduling algorithm: RANDOM. ");
      	      	
      	      	RANDOM random = new RANDOM (AortaSchedulingSimulator.ROBOT, requests, logFileWriter);
		        random.execution();
      	      }      	  	
      	  }
      	logFileWriter.println("#");
      	      	
      	
      	// unlock the available candidate robots
      	if (actionOp.q.aortaqp.lockingEnabled)
      	  for (i=0;i<didVector.size();i++)
      	    actionOp.q.aortaqp.lockRobotVector.removeElement((String)didVector.elementAt(i));
      }	
} 

