/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

/**
 * Generate a synthetic action workload with a skewed request distribution on devices.
 *
 * @see AortaSchedulingSimulator
 * @see UniformSyntheticWorkload
 * @see ZipfianSyntheticWorkload
 * @see UniformTCSyntheticWorkload
 * @see ZipfianTCSyntheticWorkload
 */
public class SkewedSyntheticWorkload {
	
    /** Type of device involved in the scheduling. */
    public byte deviceType; 
    
    /** Number of devices involved in the scheduling. */
    public int deviceNum;
    
    /** Number of action requests involved in the scheduling. */
    public int requestNum;
     
    /** A value in [0,1] that indicates the portion of the total number of cameras with a skewed workload on. */
    public double skewness; 
    
    Random ran = new Random(); 
    
                                       
	public SkewedSyntheticWorkload (byte deviceType, int deviceNum, int requestNum, double skewness) 
	  {
		this.deviceType = deviceType;
		
		this.deviceNum = deviceNum;
		
		this.requestNum = requestNum;
		
		this.skewness = skewness;
	  }
	
	
	/** 
	 * Generate the skewed workload. 
	 * Half of the requests are skewed, half are not. 
	 */
	public Vector generate () 
	  {
		int i;
		Vector requests = new Vector (30,10);		
		
		for (i=0;i<(requestNum/2);i++)
		  generateUniformRequest(requests, i+1); 
		  		
		for (;i<requestNum;i++)
		  generateSkewedRequest(requests, i+1); 		
		
		return requests;		
	  } 	  
	  
	/** Generate a single skewed action request. */  
	public void generateSkewedRequest(Vector requests, int id) 
	  {
		ActionRequest ar = new ActionRequest ();
		
		ar.aid = String.valueOf(id);			
		
	    double cost;
		int skeweddeviceNum = (int)(deviceNum * skewness);		
		for (int i=0;i<skeweddeviceNum;i++)
		  {
		   	ar.didVector.addElement(String.valueOf(i+1));
		  	
		  	cost = ranActionCost(ar);
		  	ar.costVector.addElement(String.valueOf(cost));	  	
		  } 
		
		requests.addElement((ActionRequest)ar);			
	  } 	 
		
	/** Generate a single uniform action request. */
	public void generateUniformRequest(Vector requests, int id)  
	  {
		ActionRequest ar = new ActionRequest ();
		
		ar.aid = String.valueOf(id);						
		
	    double cost;
		for (int i=0;i<deviceNum;i++)
		  {
		  	ar.didVector.addElement(String.valueOf(i+1));
		  	
		  	cost = ranActionCost(ar);
		  	ar.costVector.addElement(String.valueOf(cost));	  	
		  } 
		
		requests.addElement((ActionRequest)ar);		
      } 
       
    
    /** Randomly generate the cost of an action. */
    public double ranActionCost (ActionRequest ar)
      {
      	if (deviceType == AortaSchedulingSimulator.CAMERA)
      	  return ranPhotoActionCost(ar);
      	else
      	  if (deviceType == AortaSchedulingSimulator.ROBOT)
      	    return ranWarnActionCost(ar);
      	  else
      	    {
      	      System.err.println("Unsupported device type. Must be \"camera\" or \"robot\".");
              return Double.MAX_VALUE;
            }
      }
    
    /** Randomly generate the cost of a {@link Photo} action. */
    public double ranPhotoActionCost (ActionRequest ar)
     {
       CameraSimulator camSim = new CameraSimulator("1");
       
       int pan = ranPan(); 
       int tilt = ranTilt();
       int zoom = ranZoom(); 
       
       Vector v = new Vector(10,5);
       v.addElement(String.valueOf(pan));
       v.addElement(String.valueOf(tilt));
       v.addElement(String.valueOf(zoom));
       Tuple t = new Tuple(v);
       ar.statusVector.addElement((Tuple)t);
       
       double headMovementCost = camSim.computeHeadMovementCost(pan, tilt, zoom);
       
       return 2*camSim.STANDARD_POLLING_TIME + headMovementCost + camSim.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST; 
     }
     
    /** Randomly pick a target <code>pan</code> for the camera head. */
    public int ranPan()
	  {
	  	return CameraSimulator.MIN_PAN + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_PAN - CameraSimulator.MIN_PAN + 1);
	  }
	  
    /** Randomly pick a target <code>tilt</code> for the camera head. */
    public int ranTilt()
	  {
	  	return CameraSimulator.MIN_TILT + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_TILT - CameraSimulator.MIN_TILT + 1);
	  }
	  
    /** Randomly pick a target <code>zoom</code> for the camera head. */
    public int ranZoom()
	  {
	  	int num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 1) 
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 2)
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num == 0) return 1;
	  	else if (num == 1) return 2501; 
	         else if (num == 2) return 5001;
	              else return 9999;	  	
	  }
	  
	/** Randomly generate the cost of a {@link Warn} action. */
    public double ranWarnActionCost (ActionRequest ar)
     {
       Point loc = ranLoc();
                         
       Vector v = new Vector(10,5);
       v.addElement(String.valueOf(loc.x));
       v.addElement(String.valueOf(loc.y));
       v.addElement(String.valueOf(loc.z));
       Tuple t = new Tuple(v);
       ar.statusVector.addElement((Tuple)t);
       
       double actionCost = 0;
       actionCost += Math.sqrt(Math.pow(loc.x-ER1Simulator.HOME_LOC_X,2)+Math.pow(loc.y-ER1Simulator.HOME_LOC_Y,2))*100*ER1Simulator.MOVEPERCM_ESTIMATED_COST;
       double currentX = ER1Simulator.HOME_LOC_X, currentY = ER1Simulator.HOME_LOC_Y;
       double targetX = loc.x, targetY = loc.y;  	    
       double deltaAngle; 
       if ((currentX == targetX) && (currentY == targetY))
	      deltaAngle = 0; 
	   else
	      {
	     	if (currentX == targetX)
	     	  if (currentY < targetY)
	     	    deltaAngle = 90;
	     	  else
	     	    deltaAngle = -90;
	     	else
	     	  {
	     	  	 deltaAngle = Math.atan((targetY - currentY)/(targetX-currentX))/Math.PI*180;
	     	  	  	
	     	  	 if ((deltaAngle > 0) && (targetY < currentY))
	     	  	   deltaAngle -= 180;
	     	  	  	
	     	  	 if ((deltaAngle < 0) && (targetY > currentY))
	     	  	   deltaAngle += 180;
	     	  }	     	  	
	      }	
	   actionCost += ER1Simulator.ROTATEPERDEG_ESTIMATED_COST*deltaAngle;
       actionCost += ER1Simulator.DEFAULT_SOUND_COST;
       
       return actionCost; 
     }	
     
   /** Randomly pick a target location for the robot. */
   public Point ranLoc ()
     {
       Point loc = new Point();
       
       loc.x = ER1Simulator.MIN_LOC_X + Math.abs(ran.nextInt()) % (ER1Simulator.MAX_LOC_X - ER1Simulator.MIN_LOC_X);
       
       loc.y = ER1Simulator.MIN_LOC_Y + Math.abs(ran.nextInt()) % (ER1Simulator.MAX_LOC_Y - ER1Simulator.MIN_LOC_Y);
       
       loc.z = ER1Simulator.HOME_LOC_Z;
       
       return loc;	
     }
} 